from functools import partial

import pytest
from whenever import DateDelta
from whenever import Time
from whenever import TimeDelta
from whenever import days

from textual_timepiece._extra import LockButton
from textual_timepiece.pickers import DateRangePicker
from textual_timepiece.pickers import DateTimeDurationPicker
from textual_timepiece.pickers import DateTimeRangePicker


@pytest.fixture
def date_range_app(create_app):
    return create_app(DateRangePicker)


@pytest.fixture
def datetime_range_app(create_app):
    return create_app(DateTimeRangePicker)


@pytest.fixture
def datetime_dur_range_app(create_app):
    return create_app(DateTimeDurationPicker)


@pytest.fixture
def range_snap_compare(snap_compare):
    return partial(snap_compare, terminal_size=(85, 34))


@pytest.mark.snapshot
def test_date_range_dialog(date_range_app, range_snap_compare, freeze_time):
    async def run_before(pilot):
        date_range_app.action_focus_next()
        date_range_app.widget.query_one("#target-default-start").press()
        date_range_app.widget.query_one("#target-default-end").press()
        await pilot.press("shift+enter")

    assert range_snap_compare(date_range_app, run_before=run_before)


@pytest.mark.snapshot
def test_mini_date_range_dialog(
    date_range_app,
    range_snap_compare,
    freeze_time,
):
    async def run_before(pilot):
        date_range_app.widget.add_class("mini")
        date_range_app.action_focus_next()
        date_range_app.widget.query_one("#target-default-start").press()
        await pilot.press("shift+enter")

    assert range_snap_compare(date_range_app, run_before=run_before)


@pytest.mark.unit
async def test_date_range_lock(create_app, freeze_time):
    date_range_app = create_app(DateRangePicker(date_range=DateDelta(days=5)))

    async with date_range_app.run_test() as pilot:
        date_range_app.action_focus_next()
        date_range_app.widget.query_one("#target-default-start").press()
        await pilot.pause()
        assert date_range_app.widget.end_date == freeze_time + DateDelta(
            days=5
        )

        date_range_app.widget.end_date += days(5)
        await pilot.pause()
        assert date_range_app.widget.start_date == freeze_time + DateDelta(
            days=5
        )


@pytest.mark.unit
async def test_date_range_cap_disable(date_range_app, freeze_time):
    async with date_range_app.run_test():
        assert date_range_app.widget.start_input.disabled is False
        assert date_range_app.widget.end_input.disabled is False

        date_range_app.widget.disable_end()
        assert date_range_app.widget.start_input.disabled is False
        assert date_range_app.widget.end_input.disabled

        date_range_app.widget.disable_start()
        assert date_range_app.widget.start_input.disabled
        assert date_range_app.widget.end_input.disabled

        date_range_app.widget.disable_end(disable=False)
        assert date_range_app.widget.start_input.disabled
        assert date_range_app.widget.end_input.disabled is False


@pytest.mark.snapshot
def test_dt_range_dialog(datetime_range_app, range_snap_compare, freeze_time):
    async def run_before(pilot):
        datetime_range_app.action_focus_next()
        datetime_range_app.widget.query_one("#target-default-start").press()
        datetime_range_app.widget.query_one("#target-default-end").press()
        await pilot.pause()
        datetime_range_app.widget.end_dt = (
            datetime_range_app.widget.end_dt.add(weeks=2)
        )

        await pilot.press("shift+enter")

    assert range_snap_compare(datetime_range_app, run_before=run_before)


@pytest.mark.snapshot
def test_mini_dt_range_dialog(
    datetime_range_app, range_snap_compare, freeze_time
):
    async def run_before(pilot):
        datetime_range_app.add_class("mini")
        datetime_range_app.action_focus_next()
        datetime_range_app.widget.query_one("#target-default-start").press()

        await pilot.press("shift+enter")

    assert range_snap_compare(datetime_range_app, run_before=run_before)


@pytest.mark.snapshot
def test_dt_dur_range_dialog(
    datetime_dur_range_app, range_snap_compare, freeze_time
):
    async def run_before(pilot):
        datetime_dur_range_app.action_focus_next()
        datetime_dur_range_app.widget.query_one(
            "#target-default-start"
        ).press()
        datetime_dur_range_app.widget.query_one("#target-default-end").focus()
        await pilot.press("enter")
        datetime_dur_range_app.widget.end_dt = (
            datetime_dur_range_app.widget.end_dt.add(weeks=2)
        )
        await pilot.press("shift+enter")
        await pilot.pause()

    assert range_snap_compare(datetime_dur_range_app, run_before=run_before)


@pytest.mark.snapshot
def test_mini_dt_dur_range_dialog(
    datetime_dur_range_app, range_snap_compare, freeze_time
):
    async def run_before(pilot):
        datetime_dur_range_app.widget.add_class("mini")
        datetime_dur_range_app.action_focus_next()
        await pilot.pause()

    assert range_snap_compare(datetime_dur_range_app, run_before=run_before)


@pytest.mark.unit
async def test_dt_dur_range_lock(create_app, freeze_time):
    datetime_dur_range_app = create_app(
        DateTimeDurationPicker(time_range=TimeDelta(hours=48))
    )

    async with datetime_dur_range_app.run_test() as pilot:
        datetime_dur_range_app.action_focus_next()
        datetime_dur_range_app.widget.query_one(
            "#target-default-start"
        ).press()
        await pilot.pause()
        assert (
            datetime_dur_range_app.widget.end_date
            == freeze_time + DateDelta(days=2)
        )


@pytest.mark.unit
async def test_datetime_range_bindings(
    datetime_range_app, range_snap_compare, freeze_time
):
    async with datetime_range_app.run_test() as pilot:
        datetime_range_app.action_focus_next()

        await pilot.press("ctrl+t")
        assert datetime_range_app.widget.start_dt.date() == freeze_time

        await pilot.press("alt+ctrl+t")
        assert datetime_range_app.widget.end_dt.date() == freeze_time

        await pilot.press("ctrl+shift+d")
        assert datetime_range_app.widget.start_dt is None
        assert datetime_range_app.widget.end_dt is None


@pytest.mark.unit
async def test_dt_range_cap_disable(datetime_range_app, freeze_time):
    async with datetime_range_app.run_test():
        assert datetime_range_app.widget.start_input.disabled is False
        assert datetime_range_app.widget.end_input.disabled is False

        datetime_range_app.widget.disable_end()
        assert datetime_range_app.widget.start_input.disabled is False
        assert datetime_range_app.widget.end_input.disabled

        datetime_range_app.widget.disable_start()
        assert datetime_range_app.widget.start_input.disabled
        assert datetime_range_app.widget.end_input.disabled

        datetime_range_app.widget.disable_end(disable=False)
        assert datetime_range_app.widget.start_input.disabled
        assert datetime_range_app.widget.end_input.disabled is False


@pytest.mark.unit
async def test_datetime_range_lock(datetime_range_app, freeze_time):
    async with datetime_range_app.run_test() as pilot:
        datetime_range_app.action_focus_next()

        await pilot.press("ctrl+t")
        assert datetime_range_app.widget.start_dt.date() == freeze_time

        datetime_range_app.widget.end_dt = None
        button = datetime_range_app.widget.query_one(LockButton).press()
        await pilot.pause()
        assert button.locked is False

        datetime_range_app.widget.end_dt = freeze_time.at(Time()).add(
            days=5, ignore_dst=True
        )

        button.press()
        assert button.locked
        await pilot.pause()
        datetime_range_app.widget.end_dt = (
            datetime_range_app.widget.end_dt.add(days=5)
        )
        assert datetime_range_app.widget.start_dt.date() == freeze_time.add(
            days=5
        )



import pytest
from whenever import Time
from whenever import TimeDelta

from textual_timepiece.pickers import DurationPicker
from textual_timepiece.pickers import TimePicker
from textual_timepiece.pickers import TimeSelect


@pytest.fixture
def duration_app(create_app):
    return create_app(DurationPicker)


@pytest.fixture
def time_app(create_app):
    return create_app(TimePicker)


@pytest.mark.snapshot
def test_duration_dialog(duration_app, snap_compare, freeze_time):
    async def run_before(pilot):
        duration_app.action_focus_next()
        duration_app.widget.query_one("#target-default").press()
        await pilot.press("shift+enter")

    assert snap_compare(duration_app, run_before=run_before)


@pytest.mark.snapshot
def test_mini_duration_dialog(duration_app, snap_compare, freeze_time):
    async def run_before(pilot):
        duration_app.widget.add_class("mini")
        duration_app.action_focus_next()
        duration_app.widget.query_one("#target-default").press()
        await pilot.press("shift+enter")

    assert snap_compare(duration_app, run_before=run_before)


@pytest.mark.unit
async def test_duration_dialog_messages(duration_app, freeze_time):
    async with duration_app.run_test() as pilot:
        duration_app.action_focus_next()
        await pilot.press("shift+enter", "tab", "enter", "tab", "enter")
        assert duration_app.widget.duration.in_minutes() == 60


@pytest.mark.unit
async def test_duration_pick_spinbox(duration_app, freeze_time):
    async with duration_app.run_test() as pilot:
        duration_app.widget.query_one("#target-default").press()

        await pilot.press("up", "up")
        assert duration_app.widget.duration.in_seconds() == 1

        await pilot.press("left", "left", "left", "up")
        assert duration_app.widget.duration.in_seconds() == 61

        await pilot.press("left", "left", "down")
        assert duration_app.widget.duration == TimeDelta()


@pytest.mark.snapshot
def test_time_dialog(time_app, snap_compare, freeze_time):
    async def run_before(pilot):
        time_app.action_focus_next()
        time_app.widget.query_one("#target-default").press()
        await pilot.press("shift+enter")

    assert snap_compare(time_app, run_before=run_before)


@pytest.mark.snapshot
def test_mini_time_dialog(time_app, snap_compare, freeze_time):
    async def run_before(pilot):
        time_app.widget.add_class("mini")
        time_app.action_focus_next()
        time_app.widget.query_one("#target-default").press()
        await pilot.press("shift+enter")

    assert snap_compare(time_app, run_before=run_before)


@pytest.mark.unit
async def test_time_dialog_messages(time_app, freeze_time):
    async with time_app.run_test() as pilot:
        time_app.action_focus_next()
        await pilot.press("shift+enter", "tab", "enter", "tab", "enter")
        assert time_app.widget.time == Time(1)

        time_app.widget.query_one(TimeSelect).action_focus_neighbor("up")
        await pilot.press("up", "enter")

        assert time_app.widget.time == Time(22)


@pytest.mark.unit
async def test_time_pick_spinbox(time_app, freeze_time):
    async with time_app.run_test() as pilot:
        time_app.widget.query_one("#target-default").press()
        time_app.widget.input_widget.focus()

        await pilot.press("left", "left", "left", "up")

        assert time_app.widget.time == Time(minute=1)

        await pilot.press("right", "right", "down")
        assert time_app.widget.time == Time(second=59)



import pytest
from rich.segment import Segment
from textual.geometry import Offset
from textual.strip import Strip

from textual_timepiece._extra import BaseWidget

DATA = [
    ["what", "is", "this", "today"],
    ["who", "are", "you", "today"],
]


class LineApiWidget(BaseWidget):
    DEFAULT_CSS = """
    LineApiWidget {
        height: 2;
    }
    """

    def render_line(self, y: int):
        return Strip([Segment(t) for t in DATA[y]])


@pytest.mark.unit
def test_get_line_offset(create_app):
    app = create_app(LineApiWidget)

    assert app.widget.get_line_offset(Offset(4, 0)) == "is"

    app.widget.styles.border = ("tall", "green")

    assert app.widget.get_line_offset(Offset(5, 1)) == "is"
    assert app.widget.get_line_offset(Offset(5, 2)) == "are"



from __future__ import annotations

from abc import abstractmethod
from functools import cached_property
from typing import Callable
from typing import ClassVar
from typing import Generic
from typing import TypeVar
from typing import cast

from textual import on
from textual.binding import Binding
from textual.binding import BindingType
from textual.events import DescendantBlur
from textual.events import MouseDown
from textual.events import MouseMove
from textual.events import MouseScrollDown
from textual.events import MouseScrollUp
from textual.events import MouseUp
from textual.events import Resize
from textual.geometry import Offset
from textual.geometry import Size
from textual.reactive import var
from textual.widget import Widget
from textual.widgets import Button
from textual.widgets import Input
from textual.widgets import MaskedInput

from textual_timepiece._extra import BaseMessage
from textual_timepiece._extra import BaseWidget
from textual_timepiece._extra import ExpandButton


class BaseOverlayWidget(BaseWidget):
    """Base Class that defines the internal widgets of the dialog."""

    DEFAULT_CSS: ClassVar[str] = """
    BaseOverlayWidget {
        width: 40;
        height: auto;
    }
    """

    can_focus = True

    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        disabled: bool = False,
    ) -> None:
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)


class BaseOverlay(BaseWidget):
    """Base class for the widget that drops down for an easier selection."""

    class Close(BaseMessage):
        widget: BaseOverlay

    DEFAULT_CSS: ClassVar[str] = """
    BaseOverlay {
        overlay: screen !important;
        constrain: inflect;
        position: absolute;
        height: auto;
        width: auto;
        background: $surface;
        box-sizing: content-box;
        opacity: 0;

        border: round $secondary;

        &:focus,
        &:focus-within {
            border: round $primary;
        }

        BaseOverlayWidget {
            background: transparent;
            border: none;
        }

    }
    """

    can_focus = True

    BINDINGS: ClassVar = [
        Binding("escape", "close_dialog", "Close dialog."),
    ]

    show = var[bool](False, init=False)

    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )

        self.display = False

    def action_close_dialog(self) -> None:
        self.post_message(self.Close(self))

    def watch_show(self, show: bool) -> None:
        def anim(on_complete: Callable | None) -> None:
            self.styles.animate(
                "opacity",
                1 if show else 0,
                duration=0.4,
                easing="in_out_expo",
                on_complete=on_complete,
            )

        if show:
            self.display = show
            anim(None)
        else:
            anim(lambda: setattr(self, "display", show))

        self.set_class(show, "-expanded")

    def on_resize(self, event: Resize) -> None:
        # TODO: Need a better way to calculate this.
        # NOTE: Might have to set a constant height in a class var
        parent = cast(Widget, self.parent)
        offset = 1 if parent.has_class("mini") else 3
        bottom = self.app.size.height // 2 > cast(Widget, parent).region.y
        self.offset = Offset(0, offset if bottom else -(self.size.height + 2))

    def on_focus(self) -> None:
        self.app.action_focus_next()

    def check_action(
        self, action: str, parameters: tuple[object, ...]
    ) -> bool | None:
        if action == "focus_next_select":
            return len(self.children) > 1

        return True


T = TypeVar("T")


# TODO: Rewrite a better masked input suitable for time input.
# NOTE: Current implementation of masked input is highly restrictive. I need
# a more flexible version in order to allow for different formats to be parsed
# on the fly.
class AbstractInput(MaskedInput, BaseWidget, Generic[T]):
    """Abstract class that defines behaviour for all datetime input widgets.

    Default Input messages are disabled and are meant to be replaced by a
    custom message that returns a date/time object.

    Params:
        value: Generic value that will be placed into initialize the widget.
        name: Name of the widget.
        id: Unique DOM indentifier for the widget.
        classes: CSS classes to add.
        tooltip: Tooltip for the widget. None if not showing anything.
        disabled: Whether to disable the widget at the start.
        select_on_focus: Focus the first position on wiget focus.
        valid_empty: If the widget is valid when empty or not.
        spinbox_sensitivity: How sensitive the spinbox features are.
    """

    can_focus = True

    DEFAULT_CSS: ClassVar[str] = """
    AbstractInput {
        background: transparent;
        width: auto;
        border: none;
    }
    """

    BINDING_GROUP_TITLE = "Datetime Picker"
    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("escape", "leave", "Defocus", tooltip="Defocus the input."),
        Binding(
            "up",
            "adjust_time(1)",
            "Increment",
            tooltip="Increment value depending on keyboard cursor location.",
            priority=True,
        ),
        Binding(
            "down",
            "adjust_time(-1)",
            "Decrement",
            tooltip="Decrement value depending on keyboard cursor location.",
            priority=True,
        ),
    ]
    """All bindings for an `AbstractInput`.

    | Key(s) | Description |
    | :- | :- |
    | escape | Defocus the input. |
    | up | Increment value depending on keyboard cursor location. |
    | down | Decrement value depending on keyboard cursor location. |
    """

    ALIAS: ClassVar[str]
    PATTERN: ClassVar[str]

    def __init__(
        self,
        value: T | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        tooltip: str | None = None,
        *,
        disabled: bool = False,
        select_on_focus: bool = True,
        valid_empty: bool = True,
        spinbox_sensitivity: int = 1,
    ) -> None:
        super().__init__(
            template=self.PATTERN,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            validate_on=["changed"],
            select_on_focus=select_on_focus,
            tooltip=tooltip,
            valid_empty=valid_empty,
        )
        self.alias = value
        self._sbox_sensitivity = max(1, spinbox_sensitivity)
        self.disable_messages(Input.Changed, Input.Submitted, Input.Blurred)

    def watch_updated(self, value: bool) -> None:
        self.set_class(value, "updated")

    @abstractmethod
    def convert(self) -> T | None: ...

    def _action_leave(self) -> None:
        self.blur()

    async def _on_mouse_down(self, event: MouseDown) -> None:
        if not self.app.mouse_captured:
            self.capture_mouse()

    async def _on_mouse_move(self, event: MouseMove) -> None:
        if self.app.mouse_captured == self:
            self.action_adjust_time(-event.delta_y * self._sbox_sensitivity)

    async def _on_mouse_up(self, event: MouseUp) -> None:
        if self.app.mouse_captured == self:
            self.capture_mouse(False)

    def _on_mouse_scroll_up(self, event: MouseScrollUp) -> None:
        if self.has_focus:
            self.action_adjust_time(1 * self._sbox_sensitivity)
            event.stop()

    def _on_mouse_scroll_down(self, event: MouseScrollDown) -> None:
        if self.has_focus:
            self.action_adjust_time(-1 * self._sbox_sensitivity)
            event.stop()

    @abstractmethod
    def action_adjust_time(self, value: int) -> None: ...

    def get_content_width(self, container: Size, viewport: Size) -> int:
        return len(self.PATTERN) + 1

    @property
    def alias(self) -> T | None:
        """Alias for whatever value the input may be holding."""
        return getattr(self, self.ALIAS)

    @alias.setter
    def alias(self, value: T | None) -> None:
        """Alias for whatever value the input may be holding."""
        self.set_reactive(getattr(self.__class__, self.ALIAS), value)


Overlay = TypeVar("Overlay", bound=BaseOverlay)


class AbstractPicker(BaseWidget, Generic[Overlay]):
    """Abstract Picker class that defines most of the base behaviour."""

    DEFAULT_CSS: ClassVar[str] = """
    AbstractPicker {
        layers: base dialog;
        layout: vertical;
        height: 3;
        width: auto;

        &.mini {
            max-height: 1;
            & > #input-control {
                border: none;
                height: 1;
                padding: 0;

                &:blur {
                    padding: 0;
                }
                &:focus-within {
                    padding: 0;
                    border: none;
                }
                Button, AbstractInput {
                    border: none;
                    padding: 0;
                    height: 1;

                    &:focus {
                        color: $accent;
                        text-style: none;
                    }
                    &:disabled {
                        opacity: 50%;
                        text-style: italic;
                    }
                }
            }
        }

        & > #input-control {
            background: $surface;
            width: auto;

            &:blur {
                padding: 1;
            }
            &:focus-within {
                border: tall $primary;
                padding: 0;
            }

            Button, AbstractInput {
                border: none;
                padding: 0;
                height: 1;

                &:focus {
                    color: $accent;
                    text-style: none;
                }
            }
            & > TargetButton {
                min-width: 1;
                max-width: 3;
            }

            & > AbstractInput {
                padding: 0 2;
                &.-invalid {
                    color: $error;
                    text-style: italic;
                }
                &:focus {
                    tint: $primary 2%;
                }
            }
        }
        & > BaseOverlay {
            border: round $secondary;
            overlay: screen !important;
            constrain: inside;
            position: absolute;
            height: auto;
            width: auto;
            background: $surface;
            box-sizing: content-box;
            opacity: 0;

            &:focus,
            &:focus-within {
                border: round $primary;
            }

            & > BaseOverlayWidget {
                width: 40;
                height: auto;
            }
        }
    }
    """

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("shift+enter", "toggle('expanded')", "Open Overlay."),
    ]

    expanded = var[bool](False, init=False)
    """Whether the picker is showing its overlay or not."""

    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        disabled: bool = False,
        tooltip: str | None = None,
    ) -> None:
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self.tooltip = tooltip

    def _on_base_overlay_close(self, message: BaseOverlay.Close) -> None:
        message.stop()
        self.expanded = False

    def _compose_expand_button(self) -> ExpandButton:
        return ExpandButton(
            id="toggle-button",
            tooltip="Toggle the overlay.",
        ).data_bind(AbstractPicker.expanded)

    @on(Button.Pressed, "#toggle-button")
    def _expand_overlay(self, message: Button.Pressed) -> None:
        message.stop()
        self.expanded = not self.expanded

    @on(DescendantBlur)
    def close_overlay(self) -> None:
        if not self.has_focus_within:
            self.expanded = False

    def _watch_expanded(self, expanded: bool) -> None:
        if expanded:
            self.query_one(BaseOverlay).focus()

    @cached_property
    def overlay(self) -> Overlay:
        return cast(Overlay, self.query_exactly_one(BaseOverlay))


TI = TypeVar("TI", bound=AbstractInput)


class BasePicker(AbstractPicker, Generic[TI, T, Overlay]):
    """Base Picker class for combining various single ended widgets."""

    ALIAS: str
    INPUT: type[TI]

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding(
            "ctrl+shift+d",
            "clear",
            "Clear Value",
            tooltip="Clear the current value.",
        ),
        Binding(
            "ctrl+t",
            "target_default",
            "To Default Value",
            tooltip="Reset to the default value.",
        ),
    ]
    """All bindings for `BasePicker` classes.

    | Key(s) | Description |
    | :- | :- |
    | ctrl+shift+d | Clear the current value. |
    | ctrl+t | Reset to the default value. |
    """

    def __init__(
        self,
        value: T | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        disabled: bool = False,
        tooltip: str | None = None,
    ) -> None:
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            tooltip=tooltip,
        )
        self.value = value

    @on(Button.Pressed, "#target-default")
    def _action_target_default(
        self, message: Button.Pressed | None = None
    ) -> None:
        if message:
            message.stop()
        self.to_default()

    def action_clear(self) -> None:
        """Clear the input value."""
        self.input_widget.clear()

    @abstractmethod
    def to_default(self) -> None:
        """Behaviour when using the target-default button or action."""

    @cached_property
    def input_widget(self) -> TI:
        return self.query_exactly_one(self.INPUT)

    @property
    def value(self) -> T | None:
        """Alias for whatever value the picker may be holding."""
        return getattr(self, self.ALIAS)

    @value.setter
    def value(self, value: T | None) -> None:
        """Alias for whatever value the picker may be holding."""
        self.set_reactive(getattr(self.__class__, self.ALIAS), value)



from textual.app import App, ComposeResult
from textual_timepiece.pickers import DateRangePicker
from whenever import Date, weeks


class DateRangeApp(App[None]):

    def compose(self) -> ComposeResult:
        yield DateRangePicker(Date(2025, 2, 5), date_range=weeks(1))


if __name__ == "__main__":
    DateRangeApp().run()



from textual.app import App, ComposeResult
from textual_timepiece.pickers import TimeSelect


class TimeSelectApp(App[None]):

    def compose(self) -> ComposeResult:
        yield TimeSelect()
    

if __name__ == "__main__":
    TimeSelectApp().run()



from textual.app import App, ComposeResult
from textual_timepiece.pickers import DateTimePicker
from whenever import Date


class DateTimePickerApp(App[None]):
    
    def compose(self) -> ComposeResult:
        yield (dp := DateTimePicker())
        dp.set_reactive(DateTimePicker.expanded, True)


if __name__ == "__main__":
    DateTimePickerApp().run()



import pytest
from textual.app import App
from textual.widget import Widget
from whenever import Date
from whenever import Instant
from whenever import patch_current_time


class TestApp(App):
    def __init__(self, widget):
        super().__init__()
        if isinstance(widget, Widget):
            self.widget = widget
        else:
            self.widget = widget()

    def compose(self):
        yield self.widget


@pytest.fixture
def create_app():
    def generate_app(widget):
        return TestApp(widget)

    return generate_app


@pytest.fixture
def freeze_time():
    time = Instant.from_utc(2025, 2, 6)

    with patch_current_time(time, keep_ticking=False):
        yield Date(2025, 2, 6)



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import ClassVar
from typing import cast

from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.containers import Vertical
from textual.reactive import var
from textual.widgets import Input
from whenever import Date
from whenever import LocalDateTime
from whenever import SystemDateTime
from whenever import Time

from textual_timepiece._extra import BaseMessage
from textual_timepiece._extra import TargetButton
from textual_timepiece._utility import DateScope
from textual_timepiece._utility import round_time

from ._base_picker import AbstractInput
from ._base_picker import BaseOverlay
from ._base_picker import BasePicker
from ._date_picker import DateSelect
from ._time_picker import DurationSelect
from ._time_picker import TimeSelect


class DateTimeOverlay(BaseOverlay):
    date = var[Date | None](None, init=False)

    DEFAULT_CSS: ClassVar[str] = """
    DateTimeOverlay {
        layout: horizontal;
        max-width: 76;
        height: auto;
        Vertical {
            width: auto;
            height: auto;
        }
    }
    """

    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        disabled: bool = False,
    ) -> None:
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)

    def compose(self) -> ComposeResult:
        with Vertical():
            yield DateSelect().data_bind(date=DateTimeOverlay.date)
        with Vertical():
            yield DurationSelect()
            yield TimeSelect()
            # TODO: This should be toggleable and slide out at the user behest.

    @cached_property
    def date_select(self) -> DateSelect:
        return cast(DateSelect, self.query_one(DateSelect))


class DateTimeInput(AbstractInput[LocalDateTime]):
    """Input that combines both date and time into one."""

    @dataclass
    class DateTimeChanged(BaseMessage):
        """Sent when the datetime is changed."""

        widget: DateTimeInput
        datetime: LocalDateTime | None

    PATTERN: ClassVar[str] = r"0009-B9-99 99:99:99"
    FORMAT: ClassVar[str] = r"%Y-%m-%d %H:%M:%S"
    ALIAS = "datetime"

    datetime = var[LocalDateTime | None](None, init=False)
    """Current datetime or none if nothing is set."""

    def __init__(
        self,
        value: LocalDateTime | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        tooltip: str | None = None,
        *,
        disabled: bool = False,
        select_on_focus: bool = True,
        spinbox_sensitivity: int = 1,
    ) -> None:
        super().__init__(
            value=value,
            name=name,
            id=id,
            classes=classes,
            tooltip=tooltip,
            disabled=disabled,
            select_on_focus=select_on_focus,
            spinbox_sensitivity=spinbox_sensitivity,
        )

    def watch_datetime(self, value: LocalDateTime | None) -> None:
        with self.prevent(Input.Changed):
            if value:
                self.value = value.py_datetime().strftime(self.FORMAT)
            else:
                self.value = ""

        self.post_message(self.DateTimeChanged(self, self.datetime))

    def _watch_value(self, value: str) -> None:
        if (dt := self.convert()) is not None:
            self.datetime = dt

    def convert(self) -> LocalDateTime | None:
        try:
            return LocalDateTime.strptime(self.value, self.FORMAT)
        except ValueError:
            return None

    def action_adjust_time(self, offset: int) -> None:
        """Adjust date with an offset depending on the text cursor position."""

        try:
            if self.datetime is None:
                self.datetime = SystemDateTime.now().local()
            elif self.cursor_position < 4:
                self.datetime = self.datetime.add(
                    years=offset,
                    ignore_dst=True,
                )
            elif 5 <= self.cursor_position < 7:
                self.datetime = self.datetime.add(
                    months=offset,
                    ignore_dst=True,
                )
            elif 8 <= self.cursor_position < 10:
                self.datetime = self.datetime.add(
                    days=offset,
                    ignore_dst=True,
                )
            elif 11 <= self.cursor_position < 13:
                self.datetime = self.datetime.add(
                    hours=offset,
                    ignore_dst=True,
                )
            elif 14 <= self.cursor_position < 16:
                self.datetime = self.datetime.add(
                    minutes=offset,
                    ignore_dst=True,
                )
            else:
                self.datetime = self.datetime.add(
                    seconds=offset,
                    ignore_dst=True,
                )
        except ValueError as err:
            self.log.debug(err)
            if not str(err).endswith("out of range"):
                raise

    def insert_text_at_cursor(self, text: str) -> None:
        if not text.isdigit():
            return

        # Extra Date Filtering
        if self.cursor_position == 6 and text not in "012":
            return

        if self.cursor_position == 5 and text not in "0123":
            return

        if (
            self.cursor_position == 6
            and self.value[5] == "3"
            and text not in "01"
        ):
            return

        # Extra Time Filtering
        if self.cursor_position == 11:
            if (
                text == "2"
                and len(self.value) >= 12
                and self.value[12] not in "0123"
            ):
                self.value = self.value[:12] + "3" + self.value[13:]
            elif text not in "012":
                return

        if (
            self.cursor_position == 12
            and self.value[11] == "2"
            and text not in "0123"
        ):
            return

        if self.cursor_position in {14, 17} and text not in "012345":
            return

        super().insert_text_at_cursor(text)


class DateTimePicker(
    BasePicker[DateTimeInput, LocalDateTime, DateTimeOverlay]
):
    """Datetime picker with a date and time in one input.


    Params:
        value: Initial datetime value for the widget.
        name: Name for the widget.
        id: DOM identifier for the widget.
        classes: CSS classes for the widget
        disabled: Whether to disable the widget.
        tooltip: Tooltip to show on hover.
    """

    @dataclass
    class DateTimeChanged(BaseMessage):
        """Message sent when the datetime is updated."""

        widget: DateTimePicker
        datetime: LocalDateTime | None

    INPUT = DateTimeInput
    ALIAS = "datetime"

    datetime = var[LocalDateTime | None](None, init=False)
    """The current set datetime. Bound of to all subwidgets."""
    date = var[Date | None](None, init=False)
    """Computed date based on the datetime for the overlay."""

    def compose(self) -> ComposeResult:
        with Horizontal(id="input-control"):
            yield DateTimeInput().data_bind(DateTimePicker.datetime)
            yield TargetButton(
                id="target-default",
                tooltip="Set the datetime to now.",
            )
            yield self._compose_expand_button()

        yield (
            DateTimeOverlay().data_bind(
                date=DateTimePicker.date,
                show=DateTimePicker.expanded,
            )
        )

    def _compute_date(self) -> Date | None:
        if self.datetime:
            return self.datetime.date()
        return None

    def _watch_datetime(self, datetime: LocalDateTime | None) -> None:
        self.post_message(self.DateTimeChanged(self, datetime))

    def _on_date_select_date_changed(
        self,
        message: DateSelect.DateChanged,
    ) -> None:
        message.stop()
        if not message.date:
            return
        if self.datetime:
            self.datetime = self.datetime.time().on(message.date)
        else:
            self.datetime = message.date.at(Time())

    @on(DurationSelect.DurationRounded)
    def _round_time(self, message: DurationSelect.DurationRounded) -> None:
        message.stop()
        if self.datetime is None:
            return

        time = round_time(self.datetime.time(), message.value)
        self.datetime = self.datetime.replace_time(time)

    @on(DurationSelect.DurationAdjusted)
    def _adjust_time(self, message: DurationSelect.DurationAdjusted) -> None:
        message.stop()
        if self.datetime:
            self.datetime = self.datetime.add(message.delta, ignore_dst=True)
        else:
            self.datetime = SystemDateTime.now().local()

    @on(TimeSelect.TimeSelected)
    def _set_time(self, message: TimeSelect.TimeSelected) -> None:
        message.stop()
        if self.datetime is None:
            self.datetime = (
                SystemDateTime.now().local().replace_time(message.target)
            )
        else:
            self.datetime = self.datetime.replace_time(message.target)

    @on(DateTimeInput.DateTimeChanged)
    def _dt_input_changed(
        self, message: DateTimeInput.DateTimeChanged
    ) -> None:
        message.stop()
        with self.input_widget.prevent(DateTimeInput.DateTimeChanged):
            self.datetime = message.datetime

    def to_default(self) -> None:
        """Reset the picker datetime to the current time."""
        self.datetime = SystemDateTime.now().local()
        self.overlay.date_select.scope = DateScope.MONTH



from __future__ import annotations

import math
from calendar import day_abbr
from calendar import month_name
from collections.abc import Callable
from contextlib import suppress
from dataclasses import dataclass
from datetime import datetime
from functools import cached_property
from typing import TYPE_CHECKING
from typing import ClassVar
from typing import NamedTuple
from typing import TypeAlias
from typing import cast

from rich.segment import Segment
from rich.style import Style
from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.binding import BindingType
from textual.containers import Horizontal
from textual.events import Blur
from textual.events import Click
from textual.events import Focus
from textual.events import Leave
from textual.events import MouseMove
from textual.geometry import Offset
from textual.geometry import Size
from textual.reactive import reactive
from textual.reactive import var
from textual.strip import Strip
from textual.widgets import Button
from textual.widgets import Input
from whenever import Date
from whenever import DateDelta

from textual_timepiece._extra import BaseMessage
from textual_timepiece._extra import TargetButton
from textual_timepiece._types import Directions
from textual_timepiece._utility import DateScope
from textual_timepiece._utility import Scope
from textual_timepiece._utility import get_scope
from textual_timepiece.constants import LEFT_ARROW
from textual_timepiece.constants import RIGHT_ARROW
from textual_timepiece.constants import TARGET_ICON

from ._base_picker import AbstractInput
from ._base_picker import BaseOverlay
from ._base_picker import BaseOverlayWidget
from ._base_picker import BasePicker

if TYPE_CHECKING:
    pass


DisplayData: TypeAlias = Scope


# TODO: Month and year picker
# PERF: Region refreshing


class DateCursor(NamedTuple):
    """Cursor for keyboard navigation on the calendar dialog."""

    y: int = 0
    x: int = 0

    def confine(self, data: DisplayData) -> DateCursor:
        """Confines cursor to the current display data size."""
        y = min(len(data) + 1, self.y)
        x = min(len(data[y - 1]) - 1 if y else 3, max(self.x, 0))
        return DateCursor(y, x)

    def replace(self, **kwargs: int) -> DateCursor:
        """Create a new cursor with the supplied kwargs."""
        return self._replace(**kwargs)


class DateSelect(BaseOverlayWidget):
    """Date selection widget for selecting dates and date-ranges visually.

    Supports mouse and keyboard navigation with arrow keys.

    INFO:
        Control+Click/Enter will go back in scope with the top header.

    Params:
        start: Initial start date for the widget.
        end: Initial end date for the widget.
        name: Name of the widget.
        id: Unique dom id for the widget
        classes: Any CSS classes that should be added to the widget.
        is_range: Whether the selection is a range. Automatically true if an
            'end_date' or 'date_range' parameter is supplied.
        disabled: Whether to disable the widget.
        select_on_focus: Whether to place a keyboard cursor on widget focus.
        date_range: Whether to restrict the dates to a certain range.
            Will automatically convert to absolute values.
    """

    @dataclass
    class DateChanged(BaseMessage):
        """Message sent when the start date changed."""

        widget: DateSelect
        date: Date | None

    @dataclass
    class EndDateChanged(BaseMessage):
        """Message sent when the end date changed."""

        widget: DateSelect
        date: Date | None

    DEFAULT_CSS: ClassVar[str] = """
    DateSelect {
        background: $surface;
        width: auto;
        border: round $secondary;

        .dateselect--primary-date {
            color: $primary;
        }

        .dateselect--secondary-date {
            color: $secondary;
        }

        .dateselect--range-date {
            background: $panel-darken-3;
        }

        .dateselect--hovered-date {
            color: $accent;
            text-style: bold;
        }

        .dateselect--cursor-date {
            color: $accent;
            text-style: reverse bold;
        }

        .dateselect--start-date {
            color: $accent-lighten-3;
            text-style: italic;
        }

        .dateselect--end-date {
            color: $accent-lighten-3;
            text-style: italic;
        }
    }
    """
    """Default CSS for the `DateSelect` widget."""

    BINDING_GROUP_TITLE = "Date Select"

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding(
            "up",
            "move_cursor('up')",
            tooltip="Move the cursor up.",
        ),
        Binding(
            "right",
            "move_cursor('right')",
            tooltip="Move cursor to the right.",
        ),
        Binding(
            "down",
            "move_cursor('down')",
            tooltip="Move the cursor down.",
        ),
        Binding(
            "left",
            "move_cursor('left')",
            tooltip="Move the cursor to the left.",
        ),
        Binding(
            "enter",
            "select_cursor",
            tooltip="Navigate or select to the hovered part.",
        ),
        Binding(
            "ctrl+enter",
            "select_cursor(True)",
            tooltip="Reverse Navigate or select to the hovered part.",
        ),
    ]
    """All bindings for DateSelect

    | Key(s) | Description |
    | :- | :- |
    | up | Move the cursor up. |
    | right | Move cursor to the right. |
    | down | Move the cursor down. |
    | left | Move the cursor to the left. |
    | enter | Navigate or select to the hovered part. |
    | ctrl+enter | Reverse Navigate or select to the hovered part. |
    """

    COMPONENT_CLASSES: ClassVar[set[str]] = {
        "dateselect--start-date",
        "dateselect--end-date",
        "dateselect--cursor-date",
        "dateselect--hovered-date",  # NOTE: Only affects the foreground
        "dateselect--secondary-date",
        "dateselect--primary-date",
        "dateselect--range-date",  # NOTE: Only affects the background.
    }
    """All component classes for DateSelect.

    | Class | Description |
    | :- | :- |
    | `dateselect--cursor-date` | Color of label under the keyboard cursor. |
    | `dateselect--end-date` | Color of the selected end date if enabled. |
    | `dateselect--hovered-date` | Color of the mouse hovered date. |
    | `dateselect--primary-date` | Standard color of unselected dates. |
    | `dateselect--range-date` | Color of any dates if both end and start date\
            are selected |
    | `dateselect--secondary-date` | Color of weekdays labels in month view. |
    | `dateselect--start-date` | Color of selected start date. |
    """

    date = reactive[Date | None](None, init=False)
    """Start date. Bound to base dialog if using with a prebuilt picker."""

    date_range = var[DateDelta | None](None, init=False)
    """Constant date range in between the start and end dates."""

    end_date = reactive[Date | None](None, init=False)
    """End date for date ranges.

    Bound to base dialog if using with a prebuilt picker.
    """

    scope = var[DateScope](DateScope.MONTH)
    """Scope of the current date picker view."""

    loc = reactive[Date](Date.today_in_system_tz, init=False)
    """Current location of the date picker for navigation."""

    data = reactive[DisplayData](list, init=False, layout=True)
    """Data for displaying date info.

    Layout required as the size might differ between months.
    """

    header = reactive[str]("", init=False)
    """Navigation date header is computed dynamically."""

    cursor_offset = reactive[Offset | None](None, init=False)
    """Mouse cursor position for mouse navigation."""

    cursor = reactive[DateCursor | None](None, init=False)
    """Keyboard cursor position."""

    def __init__(
        self,
        start: Date | None = None,
        end: Date | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        is_range: bool = False,
        disabled: bool = False,
        select_on_focus: bool = True,
        date_range: DateDelta | None = None,
    ) -> None:
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self._is_range = is_range or bool(end) or bool(date_range)

        self._select_on_focus = select_on_focus

        self.set_reactive(DateSelect.date, start)
        self.set_reactive(DateSelect.end_date, end)
        self.set_reactive(DateSelect.date_range, date_range)

    def _compute_header(self) -> str:
        if self.scope == DateScope.YEAR:
            return str(self.loc.year)

        elif self.scope == DateScope.DECADE:
            start = math.floor(self.loc.year / 10) * 10
            return f"{start} <-> {start + 9}"

        elif self.scope == DateScope.CENTURY:
            start = math.floor(self.loc.year / 100) * 100
            return f"{start} <-> {start + 99}"

        return f"{month_name[self.loc.month]} {self.loc.year}"

    def _validate_date_range(
        self,
        date_range: DateDelta | None,
    ) -> DateDelta | None:
        if date_range is None:
            return None
        return abs(date_range)

    def _watch_date_range(self, new: DateDelta | None) -> None:
        if new is None:
            return
        self._is_range = True
        if self.date:
            self.end_date = self.date.add(new)
        elif self.end_date:
            self.date = self.end_date.subtract(new)

    def _watch_scope(self, scope: DateScope) -> None:
        self.data = get_scope(scope, self.loc)
        if self.cursor:
            self._find_move()

    def _watch_date(self, date: Date | None) -> None:
        self.scope = DateScope.MONTH
        if date:
            if self.date_range:
                self.end_date = date.add(self.date_range)

            self.loc = date

    def _watch_loc(self, loc: Date) -> None:
        self.data = get_scope(self.scope, loc)

        if self.cursor:
            self.cursor = self.cursor.confine(self.data)

    async def _on_mouse_move(self, event: MouseMove) -> None:
        self.cursor_offset = event.offset

    def _on_leave(self, event: Leave) -> None:
        self.cursor_offset = None

    def _on_blur(self, event: Blur) -> None:
        self.cursor = None

    def _on_focus(self, event: Focus) -> None:
        if self._select_on_focus:
            self.cursor = DateCursor()

    def _on_date_select_date_changed(
        self,
        message: DateSelect.DateChanged,
    ) -> None:
        self.date = message.date
        if self.date_range and message.date:
            self.end_date = message.date.add(self.date_range)

    def _on_date_select_end_date_changed(
        self,
        message: DateSelect.EndDateChanged,
    ) -> None:
        self.end_date = message.date
        if self.date_range and message.date:
            self.date = message.date.subtract(self.date_range)

    async def _on_click(self, event: Click) -> None:
        target = self.get_line_offset(event.offset)
        self._navigate_picker(target, ctrl=event.ctrl)

    def action_move_cursor(self, direction: Directions) -> None:
        """Move cursor to the next spot depending on direction."""
        if self.cursor is None:
            self.log.debug("Cursor does not exist. Placing default location.")
            self.cursor = DateCursor()
        elif direction == "up":
            self._find_move(y=-1)
        elif direction == "right":
            self._find_move(x=1)
        elif direction == "down":
            self._find_move(y=1)
        elif direction == "left":
            self._find_move(x=-1)

    def _find_move(self, *, y: int = 0, x: int = 0) -> None:
        cursor = cast(DateCursor, self.cursor)
        if (new_y := cursor.y + y) == 0:
            new_x = cursor.x + x
            if cursor.y != 0:
                # NOTE: Making sure different row lengths align.
                new_x = math.ceil(((cursor.x) / len(self.data[0])) * 3)

            self.cursor = cursor.replace(y=new_y, x=new_x).confine(self.data)

        elif y and 0 <= new_y <= len(self.data):
            new_x = cursor.x
            if cursor.y == 0:
                # NOTE: Making sure different row lengths align.
                new_x = math.ceil(((cursor.x) / 3) * len(self.data[0]))

            self.cursor = cursor.replace(y=new_y, x=new_x).confine(self.data)

        elif x and 0 <= (new_x := cursor.x + x) < len(self.data[cursor.y - 1]):
            self.cursor = cursor.replace(x=new_x).confine(self.data)

    def _set_date(self, target: str | int, *, ctrl: bool) -> None:
        try:
            value = int(target)
            date = min(Date.MAX, max(Date.MIN, self.loc.replace(day=value)))
        except ValueError:
            return
        if ctrl:
            self.post_message(self.EndDateChanged(self, date))
        else:
            self.post_message(self.DateChanged(self, date))

    def _set_month(self, target: str) -> None:
        try:
            month_no = list(month_name).index(target)
        except IndexError:
            return
        else:
            self.set_reactive(
                DateSelect.loc,
                Date(self.loc.year, month_no, self.loc.day),
            )
            self.scope = DateScope.MONTH

    def _set_years(self, target: str | int) -> None:
        if self.scope == DateScope.CENTURY and isinstance(target, str):
            target = target.split("-")[0]
        try:
            value = max(1, min(int(target), 9999))
        except ValueError:
            return
        else:
            self.set_reactive(DateSelect.loc, self.loc.replace(year=value))
            self.scope = DateScope(self.scope.value - 1)

    def _set_target(self, target: str | int, *, ctrl: bool = False) -> None:
        if self.scope == DateScope.MONTH:
            self._set_date(target, ctrl=ctrl)
        elif self.scope == DateScope.YEAR:
            self._set_month(cast(str, target))
        else:
            self._set_years(target)

    def check_action(
        self, action: str, parameters: tuple[object, ...]
    ) -> bool | None:
        if action == "select_cursor":
            return self.cursor is not None

        return True

    def action_select_cursor(self, ctrl: bool = False) -> None:
        """Triggers the functionality for what is below the keyboard cursor."""
        cursor = cast(DateCursor, self.cursor)
        if cursor.y == 0:
            nav = (
                LEFT_ARROW,
                self.header,
                TARGET_ICON,
                RIGHT_ARROW,
            )
            self._navigate_picker(nav[cursor.x], ctrl=ctrl)
        else:
            self._navigate_picker(self.data[cursor.y - 1][cursor.x], ctrl=ctrl)

    def _navigate_picker(self, target: str | int, *, ctrl: bool) -> None:
        if target == LEFT_ARROW:
            self._crement_scope(-1)
        elif target == TARGET_ICON:
            self._set_current_scope()
        elif target == RIGHT_ARROW:
            self._crement_scope(1)
        elif target == self.header:
            if ctrl:
                self.scope = DateScope(max(self.scope.value - 1, 1))
            else:
                self.scope = DateScope(min(self.scope.value + 1, 4))
        elif target or isinstance(target, int):
            self._set_target(target, ctrl=ctrl and self._is_range)

    def _set_current_scope(self) -> None:
        self.scope = DateScope.MONTH
        self.loc = self.date or self.end_date or Date.today_in_system_tz()

    def _crement_scope(self, value: int) -> None:
        with suppress(ValueError):  # NOTE: Preventing out of range values.
            if self.scope == DateScope.MONTH:
                self.loc = self.loc.add(months=value)
            elif self.scope == DateScope.YEAR:
                self.loc = self.loc.add(years=value)
            elif self.scope == DateScope.DECADE:
                self.loc = self.loc.add(years=10 * value)
            else:
                self.loc = self.loc.add(years=100 * value)

    def _filter_style(
        self,
        y: int,
        x: range,
        date: Date | None = None,
        log_idx: DateCursor | None = None,
    ) -> Style:
        """Filters a rich style based on location data.

        Args:
            y: Current row being rendered.
            x: Range of indexes to target.
            date: If a date is being filtered.
            log_idx: Logical index for rendering the keyboard cursor.

        Returns:
            Combined style with all the properties that matched.
        """
        styles = [self.get_component_rich_style("dateselect--primary-date")]

        if date:
            if date == self.date:
                styles.append(
                    self.get_component_rich_style("dateselect--start-date")
                )
            elif date == self.end_date:
                styles.append(
                    self.get_component_rich_style("dateselect--end-date")
                )

            if self.is_day_in_range(date):
                styles.append(
                    self.get_component_rich_style(
                        "dateselect--range-date"
                    ).background_style
                )

        if (
            self.cursor_offset
            and self.cursor_offset.y == y
            and self.cursor_offset.x in x
        ):
            style = self.get_component_rich_style("dateselect--hovered-date")
            styles.append(style.from_color(style.color))

        if self.cursor and self.cursor == log_idx:
            styles.append(
                self.get_component_rich_style("dateselect--cursor-date")
            )

        return Style.combine(styles)

    def is_day_in_range(self, day: Date) -> bool:
        """Checks if a given date is within selected the date range(inclusive).

        Args:
            day: Date to check against.

        Returns:
            True if in the range else false.
        """
        return bool(
            self._is_range
            and self.date
            and self.end_date
            and self.date <= day <= self.end_date
        )

    def _render_header(self, y: int) -> list[Segment]:
        header_len = len(self.header)
        rem = self.size.width - (header_len + 10)
        blank, blank_extra = divmod(rem, 2)
        header_start = 5 + blank + blank_extra
        header_end = header_start + header_len
        right_nav_start = header_end + (blank - blank_extra) + len(TARGET_ICON)

        y += self._top_border_offset()
        return [
            Segment("   ", self.rich_style),
            Segment(
                LEFT_ARROW,
                self._filter_style(
                    y,
                    range(4, 5),
                    log_idx=DateCursor(0, 0),
                ),
            ),
            Segment(" " * (blank), self.rich_style),
            Segment(
                self.header,
                style=self._filter_style(
                    y,
                    range(header_start, header_end),
                    log_idx=DateCursor(0, 1),
                ),
            ),
            Segment("   ", self.rich_style),
            Segment(
                TARGET_ICON,
                style=self._filter_style(
                    y,
                    range(header_end + 1, header_end + 3),
                    log_idx=DateCursor(0, 2),
                ),
            ),
            Segment(" " * (blank - (3 - blank_extra)), self.rich_style),
            Segment(
                RIGHT_ARROW,
                style=self._filter_style(
                    y,
                    range(right_nav_start, right_nav_start + 2),
                    log_idx=DateCursor(0, 3),
                ),
            ),
        ]

    def _render_weekdays(self) -> list[Segment]:
        day_style = self.get_component_rich_style("dateselect--secondary-date")
        empty = Segment("  ", style=self.rich_style)
        segs = [Segment(" ", style=self.rich_style)]
        for i in range(7):
            segs.append(empty)
            segs.append(Segment(day_abbr[i], day_style))
        return segs

    def _render_month(self, y: int) -> list[Segment]:
        border_offset = self._top_border_offset()
        y += border_offset
        if y == (3 + border_offset):
            return self._render_weekdays()

        month = (y - (4 + border_offset)) // 2
        # NOTE: Removing nav header + weekdays

        date = None
        segments = [Segment(" ", style=self.rich_style)]
        subtotal = int(self.styles.border_left[0] != "")
        for i in range(7):
            segments.append(
                Segment(
                    "  ",
                    self._filter_style(
                        y,
                        range(subtotal, subtotal + 3),
                        date=date,
                    ),
                )
            )
            subtotal += 2
            if not (day := self.data[month][i]):
                segments.append(
                    Segment(
                        "   ",
                        style=self._filter_style(
                            y,
                            range(subtotal, subtotal + 4),
                            date=date,
                            log_idx=DateCursor(month + 1, i),
                        ),
                    )
                )
                date = None
            else:
                date = self.loc.replace(day=cast(int, day))
                segments.append(
                    Segment(
                        str(day).rjust(3),
                        style=self._filter_style(
                            y,
                            range(subtotal, subtotal + 4),
                            date=date,
                            log_idx=DateCursor(month + 1, i),
                        ),
                    )
                )
            subtotal += 3

        return segments

    def _render_year(self, y: int) -> list[Segment]:
        if (row := (y - 2) // 2) > 3:
            return []

        y += self._top_border_offset()

        values = self.data[row]
        value_max_width = self.size.width // len(values)

        segs = list[Segment]()
        for i, value in enumerate(values):
            if self.scope == DateScope.CENTURY:
                value = f"{value}-{cast(int, value) + 9}"
            else:
                value = str(value)
            n = len(value)
            start = (i * value_max_width) + (abs(value_max_width - n) // 2)
            end = start + n + 1

            value = value.center(value_max_width)
            segs.append(
                Segment(
                    value,
                    self._filter_style(
                        y,
                        range(start, end),
                        log_idx=DateCursor(row + 1, i),
                    ),
                )
            )

        return segs

    def render_line(self, y: int) -> Strip:
        if (y % 2 == 0) or (len(self.data) + 2) * 2 < y or not self.data:
            return Strip.blank(self.size.width)

        if y == 1:
            line = self._render_header(y)
        elif self.scope == DateScope.MONTH:
            line = self._render_month(y)
        else:
            line = self._render_year(y)

        return Strip(line)

    def get_content_height(
        self,
        container: Size,
        viewport: Size,
        width: int,
    ) -> int:
        total = 3 + len(self.data) * 2

        if self.scope == DateScope.MONTH:
            return total + 2

        return total

    def get_content_width(self, container: Size, viewport: Size) -> int:
        return 38


class EndDateSelect(DateSelect):
    """Date select widget that inverts the widgets controls to select end_date
    first.

    Params:
        start: Initial start date for the widget.
        end: Initial end date for the widget.
        name: Name of the widget.
        id: Unique dom id for the widget
        classes: Any dom classes that should be added to the widget.
        disabled: Whether to disable the widget.
        select_on_focus: Whether to place a keyboard cursor on widget focus.
        date_range: Whether to restrict the dates to a concrete range.
    """

    def __init__(
        self,
        start: Date | None = None,
        end: Date | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        disabled: bool = False,
        select_on_focus: bool = True,
        date_range: DateDelta | None = None,
    ) -> None:
        super().__init__(
            start=start,
            end=end,
            name=name,
            id=id,
            classes=classes,
            is_range=True,
            disabled=disabled,
            select_on_focus=select_on_focus,
            date_range=date_range,
        )

    def _watch_end_date(self, date: Date | None) -> None:
        if date:
            self.scope = DateScope.MONTH
            self.loc = date
            if self.date_range:
                self.date = date - self.date_range

    def _set_date(self, target: str | int, *, ctrl: bool) -> None:
        try:
            value = int(target)
        except ValueError:
            return
        else:
            date = min(Date.MAX, max(Date.MIN, self.loc.replace(day=value)))
            if not ctrl:
                self.post_message(self.EndDateChanged(self, date))
            else:
                self.post_message(self.DateChanged(self, date))

    def _set_current_scope(self) -> None:
        self.set_reactive(DateSelect.scope, DateScope.MONTH)
        if self.end_date:
            self.loc = self.end_date
        elif self.date:
            self.loc = self.date


class DateOverlay(BaseOverlay):
    date = var[Date | None](None, init=False)

    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        date_range: DateDelta | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self._date_range = date_range

    def compose(self) -> ComposeResult:
        yield DateSelect(date_range=self._date_range).data_bind(
            date=DateOverlay.date
        )

    @cached_property
    def date_select(self) -> DateSelect:
        return cast(DateSelect, self.query_one(DateSelect))


class EndDateOverlay(DateOverlay):
    date = var[Date | None](None, init=False)

    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        disabled: bool = False,
    ) -> None:
        super().__init__(name, id, classes, disabled=disabled)

    def compose(self) -> ComposeResult:
        yield EndDateSelect().data_bind(end_date=EndDateOverlay.date)


# TODO: Support for End Date in same input "YYYY/MM/DD - YYYY/MM/DD"


class DateInput(AbstractInput[Date]):
    """Date picker for full dates.

    Params:
        day: Initial value to set.
        name: Name of the widget.
        id: Unique dom identifier value.
        classes: Any dom classes to add.
        tooltip: Tooltip to show when hovering the widget.
        disabled: Whether to disable the widget.
        select_on_focus: Whether to place the cursor on focus.
        spinbox_sensitivity: Sensitivity setting for spinbox functionality.
    """

    @dataclass
    class DateChanged(BaseMessage):
        """Message sent when the date changed."""

        widget: DateInput
        date: Date | None

    PATTERN: ClassVar[str] = "0009-B9-99"
    DATE_FORMAT: ClassVar[str] = "%Y-%m-%d"
    ALIAS = "date"
    date = var[Date | None](None, init=False)
    """Date that is set. Bound if using within a picker."""

    def __init__(
        self,
        day: Date | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        tooltip: str | None = "YYYY-MM-DD Format.",
        *,
        disabled: bool = False,
        select_on_focus: bool = True,
        spinbox_sensitivity: int = 1,
    ) -> None:
        super().__init__(
            value=day,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            tooltip=tooltip,
            select_on_focus=select_on_focus,
            spinbox_sensitivity=spinbox_sensitivity,
        )

    def watch_date(self, new: Date | None) -> None:
        # FIX: probably should prevent date changes
        with self.prevent(Input.Changed):
            self.value = (
                new.py_date().strftime(self.DATE_FORMAT) if new else ""
            )
        self.post_message(self.DateChanged(self, new))

    def _watch_value(self, value: str) -> None:
        if date := self.convert():
            self.date = date

    def action_adjust_time(self, offset: int) -> None:
        """Adjust date with an offset depending on the text cursor position."""

        try:
            if self.date is None:
                self.date = Date.today_in_system_tz()
            elif self._is_year_pos():
                self.date = self.date.add(years=offset)
            elif self._is_month_pos():
                self.date = self.date.add(months=offset)
            else:
                self.date = self.date.add(days=offset)
        except ValueError as err:
            self.log.debug(err)
            if not str(err).endswith("out of range"):
                raise

    def _is_year_pos(self) -> bool:
        return self.cursor_position < 4

    def _is_month_pos(self) -> bool:
        return 5 <= self.cursor_position < 7

    def _is_day_pos(self) -> bool:
        return 8 <= self.cursor_position

    def convert(self) -> Date | None:
        # NOTE: Pydate instead as I want to keep it open to standard formats.
        try:
            return Date.from_py_date(
                datetime.strptime(self.value, self.DATE_FORMAT).date()
            )
        except ValueError:
            return None

    def insert_text_at_cursor(self, text: str) -> None:
        if not text.isdigit():
            return

        # Extra Date Filtering
        if self.cursor_position == 6 and text not in "012":
            return

        if self.cursor_position == 5 and text not in "0123":
            return

        if (
            self.cursor_position == 6
            and self.value[5] == "3"
            and text not in "01"
        ):
            return

        super().insert_text_at_cursor(text)


class DatePicker(BasePicker[DateInput, Date, DateOverlay]):
    """Single date picker with an input and overlay.

    Params:
        date: Initial date for the picker.
        name: Name for the widget.
        id: DOM identifier for widget.
        classes: Classes to add the widget.
        date_range: Date range to lock the widget to.
        disabled: Disable the widget.
        validator: A callable that will validate and adjust the date if needed.
        tooltip: A tooltip to show when hovering the widget.

    Examples:
        >>> def limit_dates(date: Date | None) -> Date | None:
        >>>     if date is None:
        >>>         return None
        >>>     return min(Date(2025, 2, 20), max(Date(2025, 2, 6), date))
        >>> yield DatePicker(validator=limit_dates)

        >>> yield DatePicker(
        >>>     Date.today_in_system_tz(),
        >>>     date_range=DateDelta(days=5),
        >>> )
    """

    @dataclass
    class DateChanged(BaseMessage):
        """Message sent when the date changed."""

        widget: DatePicker
        date: Date | None

    BINDING_GROUP_TITLE = "Date Picker"
    ALIAS = "date"

    DateValidator: TypeAlias = Callable[[Date | None], Date | None]
    """Callable type for validating date types."""

    date = var[Date | None](None, init=False)
    """Current date for the picker. This is bound to every other subwidget."""

    def __init__(
        self,
        date: Date | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        date_range: DateDelta | None = None,
        disabled: bool = False,
        validator: DateValidator | None = None,
        tooltip: str | None = None,
    ) -> None:
        super().__init__(
            value=date,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            tooltip=tooltip,
        )
        self._date_range = date_range
        self.validator = validator

    def _validate_date(self, date: Date | None) -> Date | None:
        if self.validator is None:
            return date

        return self.validator(date)

    def check_action(
        self, action: str, parameters: tuple[object, ...]
    ) -> bool | None:
        if action == "target_today":
            return self.date != Date.today_in_system_tz()
        return True

    def compose(self) -> ComposeResult:
        with Horizontal(id="input-control"):
            yield DateInput(id="date-input").data_bind(date=DatePicker.date)

            yield TargetButton(
                id="target-default",
                disabled=self.date == Date.today_in_system_tz(),
                tooltip="Set the date to today.",
            )
            yield self._compose_expand_button()

        yield (
            DateOverlay(date_range=self._date_range).data_bind(
                date=DatePicker.date, show=DatePicker.expanded
            )
        )

    def _on_date_select_date_changed(
        self,
        message: DateSelect.DateChanged,
    ) -> None:
        message.stop()
        self.date = message.date

    def _watch_date(self, new: Date) -> None:
        self.query_exactly_one("#target-default", Button).disabled = (
            new == Date.today_in_system_tz()
        )
        self.post_message(self.DateChanged(self, new))

    @on(DateInput.DateChanged)
    def _input_updated(self, message: DateInput.DateChanged) -> None:
        message.stop()
        with message.widget.prevent(DateInput.DateChanged):
            self.date = message.date

    def action_clear(self) -> None:
        """Clear the date."""
        self.date = None

    def to_default(self) -> None:
        """Reset the date to today."""
        self.overlay.date_select.scope = DateScope.MONTH
        self.date = Date.today_in_system_tz()



"""Constants that the widgets use. Mostly different unicode symbols.

Exposed as a public module in order to allow the user patch before importing
one of the widgets or setting them through environment variables.

Examples:
    >>> import textual_timepiece.constants
    >>> textual_timepiece.constants.LEFT_ARROW = "L"
    >>> textual_timepiece.constants.RIGHT_ARROW = "R"
    >>> from textual_timepiece.pickers import DatePicker
"""

import os

LEFT_ARROW: str = os.environ.get("TTIME_ICON_LEFT_ARROW", "←")
"""Left pointing symbol for navigational purposes. Mostly going back in time.
"""

RIGHT_ARROW: str = os.environ.get("TTIME_ICON_RIGHT_ARROW", "→")
"""Right pointing symbol for navigational purposes. Mostly going forward in
time.
"""

TARGET_ICON: str = os.environ.get("TTIME_ICON_TARGET", "◎")
"""Used as the default target icon. Mostly when going to the current
date and/or time.
"""

LOCKED_ICON: str = os.environ.get("TTIME_ICON_LOCKED", "🔒")
"""Icon that is rendered when a widget is locked."""

UNLOCKED_ICON: str = os.environ.get("TTIME_ICON_UNLOCKED", "🔓")
"""Icon that is rendered when a widget is unlocked."""



from textual.app import App, ComposeResult
from textual_timepiece.pickers import TimePicker
from whenever import Date, Time


class TimePickerApp(App[None]):
    
    def compose(self) -> ComposeResult:
        yield (dp := TimePicker(Time(12, 0, 0)))
        dp.set_reactive(TimePicker.expanded, True)


if __name__ == "__main__":
    TimePickerApp().run()



from ._date_picker import DateInput
from ._date_picker import DateOverlay
from ._date_picker import DatePicker
from ._date_picker import DateSelect
from ._date_picker import EndDateOverlay
from ._date_picker import EndDateSelect
from ._datetime_picker import DateTimeInput
from ._datetime_picker import DateTimeOverlay
from ._datetime_picker import DateTimePicker
from ._time_picker import DurationInput
from ._time_picker import DurationOverlay
from ._time_picker import DurationPicker
from ._time_picker import DurationSelect
from ._time_picker import TimeInput
from ._time_picker import TimeOverlay
from ._time_picker import TimePicker
from ._time_picker import TimeSelect
from ._timerange_picker import DateRangePicker
from ._timerange_picker import DateTimeDurationPicker
from ._timerange_picker import DateTimeRangePicker

__all__ = [
    "DateInput",
    "DateOverlay",
    "DatePicker",
    "DateRangePicker",
    "DateSelect",
    "DateTimeDurationPicker",
    "DateTimeInput",
    "DateTimeOverlay",
    "DateTimePicker",
    "DateTimeRangePicker",
    "DurationInput",
    "DurationOverlay",
    "DurationPicker",
    "DurationSelect",
    "EndDateOverlay",
    "EndDateSelect",
    "TimeInput",
    "TimeOverlay",
    "TimePicker",
    "TimeSelect",
]



"""Activity Heatmap graph widgets for a yearly overview."""

from ._activity_heatmap import ActivityHeatmap
from ._activity_heatmap import HeatmapManager

__all__ = ("ActivityHeatmap", "HeatmapManager")



import pytest
from whenever import Date
from whenever import LocalDateTime
from whenever import Time

from textual_timepiece.pickers import DateSelect
from textual_timepiece.pickers import DateTimePicker
from textual_timepiece.pickers import TimeSelect


@pytest.fixture
def datetime_app(create_app):
    return create_app(DateTimePicker)


@pytest.mark.unit
async def test_datetime_default(datetime_app, freeze_time):
    async with datetime_app.run_test() as pilot:
        datetime_app.action_focus_next()
        datetime_app.widget.query_one("#target-default").focus()
        await pilot.press("enter")
        assert datetime_app.widget.datetime == freeze_time.at(Time())


@pytest.mark.snapshot
def test_datetime_dialog(datetime_app, snap_compare, freeze_time):
    async def run_before(pilot):
        datetime_app.action_focus_next()
        datetime_app.widget.query_one("#target-default").press()
        await pilot.press("shift+enter")

    assert snap_compare(datetime_app, run_before=run_before)


@pytest.mark.snapshot
def test_mini_datetime_dialog(datetime_app, snap_compare, freeze_time):
    async def run_before(pilot):
        datetime_app.widget.add_class("mini")
        datetime_app.action_focus_next()
        datetime_app.widget.query_one("#target-default").press()
        await pilot.press("shift+enter")

    assert snap_compare(datetime_app, run_before=run_before)


@pytest.mark.unit
async def test_datetime_dialog_messages(datetime_app, freeze_time):
    async with datetime_app.run_test() as pilot:
        datetime_app.action_focus_next()
        datetime_app.widget.query_one("#target-default").focus()

        await pilot.press("shift+enter", "down", "down", "enter")
        assert datetime_app.widget.datetime.date() == Date(2025, 2, 3)

        datetime_app.widget.query_one(TimeSelect).action_focus_neighbor("up")
        await pilot.press("up", "enter")
        assert datetime_app.widget.datetime == Date(2025, 2, 3).at(Time(22))


@pytest.mark.unit
async def test_datetime_dialog_messages_reverse(datetime_app, freeze_time):
    async with datetime_app.run_test() as pilot:
        datetime_app.action_focus_next()

        await pilot.press("shift+enter")
        datetime_app.widget.query_one(TimeSelect).action_focus_neighbor("up")
        await pilot.press("up", "enter")
        assert datetime_app.widget.datetime == freeze_time.at(Time(22))
        datetime_app.widget.query_one("#target-default").focus()

        datetime_app.widget.query_one(DateSelect).focus()
        await pilot.press("down", "down", "enter")
        assert datetime_app.widget.datetime == Date(2025, 2, 3).at(Time(22))


@pytest.mark.unit
async def test_dt_pick_spinbox(datetime_app, freeze_time):
    async with datetime_app.run_test() as pilot:
        datetime_app.widget.query_one("#target-default").press()
        datetime_app.widget.input_widget.focus()

        await pilot.press("right", "right", "right", "up")
        assert datetime_app.widget.datetime == freeze_time.replace(
            year=2026
        ).at(Time())

        await pilot.press("right", "right", "down")
        assert datetime_app.widget.datetime == freeze_time.replace(
            year=2026, month=1
        ).at(Time())

        datetime_app.widget.input_widget.cursor_position = 0
        await pilot.press("3")
        assert datetime_app.widget.datetime == freeze_time.replace(
            year=3026, month=1
        ).at(Time())

        await pilot.press("shift+end", "left", "down")
        assert datetime_app.widget.datetime == freeze_time.replace(
            year=3026, month=1, day=5
        ).at(Time(23, 59, 59))


@pytest.mark.unit
async def test_dt_dialog_edge_cases(datetime_app, freeze_time):
    async with datetime_app.run_test() as pilot:
        datetime_app.widget.focus()
        select = datetime_app.widget.overlay.date_select
        select.post_message(DateSelect.DateChanged(select, Date.MIN))
        await pilot.pause()
        assert datetime_app.widget.date == Date.MIN

        select.post_message(DateSelect.DateChanged(select, Date.MAX))
        await pilot.pause()
        assert datetime_app.widget.date == Date.MAX


@pytest.mark.unit
async def test_dt_input_edge_cases(datetime_app, freeze_time):
    async with datetime_app.run_test() as pilot:
        datetime_app.widget.input_widget.focus()
        datetime_app.widget.datetime = LocalDateTime.MAX
        await pilot.press("up")

        datetime_app.widget.datetime = LocalDateTime.MIN
        await pilot.press("down")



from __future__ import annotations

import inspect
from dataclasses import dataclass
from functools import cached_property
from typing import ClassVar
from typing import Literal

from rich.console import RenderableType
from rich.pretty import Pretty
from rich.syntax import Syntax
from textual import on
from textual.app import App
from textual.app import ComposeResult
from textual.containers import Container
from textual.containers import Horizontal
from textual.containers import ScrollableContainer
from textual.message import Message
from textual.screen import ModalScreen
from textual.widget import Widget
from textual.widgets import Button
from textual.widgets import Footer
from textual.widgets import Header
from textual.widgets import Label
from textual.widgets import Static
from textual.widgets import TabbedContent
from textual.widgets import TabPane

from textual_timepiece.__about__ import __version__
from textual_timepiece._activity_heatmap import ActivityHeatmap
from textual_timepiece._activity_heatmap import HeatmapManager
from textual_timepiece.pickers import DatePicker
from textual_timepiece.pickers import DateRangePicker
from textual_timepiece.pickers import DateTimeDurationPicker
from textual_timepiece.pickers import DateTimePicker
from textual_timepiece.pickers import DateTimeRangePicker
from textual_timepiece.pickers import DurationPicker
from textual_timepiece.pickers import TimePicker
from textual_timepiece.pickers._date_picker import DateSelect
from textual_timepiece.pickers._time_picker import DurationSelect
from textual_timepiece.pickers._time_picker import TimeSelect


class DemoWidget(Widget):
    @dataclass
    class ToggleFeature(Message):
        widget: type[Widget]
        preview: Literal[
            "docstring", "tcss", "code", "docs", "source", "bindings"
        ]

    def __init__(
        self,
        widget_call: type[Widget],
        *,
        notes: str | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)

        self._notes = notes
        self._widget_type = widget_call

    def _compose_navigation_bar(self) -> ComposeResult:
        with Horizontal(id="navigation"):
            yield Button("Docstring", id="docstring", classes="nav")
            yield Label(self._widget_type.__name__, classes="title")
            if (
                hasattr(self._widget_type, "DEFAULT_CSS")
                and self._widget_type.DEFAULT_CSS
            ):
                yield Button("Default CSS", id="default-css", classes="nav")
            if (
                hasattr(self._widget_type, "BINDINGS")
                and self._widget_type.BINDINGS
            ):
                yield Button("Bindings", id="bindings", classes="nav")

            yield Button("Code Preview", id="code", classes="nav")

    def compose(self) -> ComposeResult:
        yield from self._compose_navigation_bar()
        yield self._widget_type()

    @on(Button.Pressed, "#docstring")
    def open_docstring(self, message: Button.Pressed) -> None:
        self.post_message(self.ToggleFeature(self._widget_type, "docstring"))

    @on(Button.Pressed, "#default-css")
    def open_default_css(self, message: Button.Pressed) -> None:
        self.post_message(self.ToggleFeature(self._widget_type, "tcss"))

    @on(Button.Pressed, "#code")
    def open_source(self, message: Button.Pressed) -> None:
        self.post_message(self.ToggleFeature(self._widget_type, "code"))

    @on(Button.Pressed, "#bindings")
    def open_bindings(self, message: Button.Pressed) -> None:
        self.post_message(self.ToggleFeature(self._widget_type, "bindings"))


class PreviewScreen(ModalScreen[None]):
    BINDINGS: ClassVar = [
        ("escape", "hide_preview", "Close Preview"),
    ]

    def __init__(
        self,
        renderable: RenderableType,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        super().__init__(name, id, classes)
        self._renderable = renderable

    def on_mount(self) -> None:
        self.refresh_bindings()

    def compose(self) -> ComposeResult:
        with Container():
            with ScrollableContainer():
                yield Static(self._renderable, id="preview")

            with Horizontal():
                yield Button(
                    r"Quit\[esc]",
                    "warning",
                    action="screen.hide_preview",
                    classes="nav",
                )

    def action_hide_preview(self) -> None:
        self.dismiss()


class TimepieceDemo(App[None]):
    font: bool = True

    CSS = """
    Screen {
        layout: horizontal;
        align: center middle;

        TabbedContent {
            padding: 1 0;
            min-width: 135;
            width: 65%;

            TabPane {
                height: auto;
            }

            Container.previews {
                layout: vertical;
                overflow-y: auto;
                content-align-horizontal: center;
                align-horizontal: center;
            }
        }
    }

    .title {
        background: $panel;
        content-align-horizontal: center;
        text-align: center;
        width: 1fr;
        text-style: bold;
    }

    DemoWidget {
        height: auto;
        margin: 2 0;
        padding: 0 2;

        Horizontal#navigation {
            height: auto;
            margin-bottom: 1;
        }
    }

    Button.nav {
        border: none;
        height: 1;
        width: 10%;
        min-width: 13;
    }

    .widget_container {
        height: auto;
        width: 100%;
        align: center middle;
    }

    PreviewScreen {
        align: center middle;

        Container {
            border: round $primary;
            width: 50%;
            min-width: 125;
            height: 90%;

            ScrollableContainer {
                Static#preview {
                    margin: 1 2;
                }
            }

            Horizontal {
                height: 1;
                margin-top: 1;
                align-horizontal: center;
            }

        }
    }

    """

    TITLE = "Textual Timepiece"
    SUB_TITLE = __version__

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)

        with TabbedContent(initial="pickers"):
            with TabPane("Pickers", id="pickers"):
                with Container(id="Pickers", classes="previews"):
                    for item in (
                        DatePicker,
                        DurationPicker,
                        TimePicker,
                        DateTimePicker,
                        DateRangePicker,
                        DateTimeRangePicker,
                        DateTimeDurationPicker,
                    ):
                        yield DemoWidget(item)

            with TabPane("Select"):
                with Container(id="Pickers", classes="previews"):
                    for select in (
                        DateSelect,
                        TimeSelect,
                        DurationSelect,
                    ):
                        yield DemoWidget(select)

            with TabPane("Heatmap"):
                with Container(id="heatmap", classes="previews"):
                    for i in (ActivityHeatmap, HeatmapManager):
                        yield DemoWidget(i)

        yield Footer()

    def on_mount(self) -> None:
        fake_data = ActivityHeatmap.generate_empty_activity(2025)

        for m in self.query(ActivityHeatmap):
            m.process_data(fake_data)

    @on(DemoWidget.ToggleFeature)
    def open_tab(self, message: DemoWidget.ToggleFeature) -> None:
        data: RenderableType
        if message.preview == "tcss":
            data = message.widget.DEFAULT_CSS

        elif message.preview == "bindings":
            data = Pretty(message.widget.BINDINGS)

        elif message.preview == "docstring":
            data = str(message.widget.__doc__)

        elif message.preview == "code":
            data = Syntax(
                inspect.getsource(message.widget),
                "python",
                line_numbers=True,
                padding=1,
            )

        self.app.push_screen(PreviewScreen(data))

    @on(HeatmapManager.YearChanged)
    def change_heat_year(self, message: HeatmapManager.YearChanged) -> None:
        fake_data = ActivityHeatmap.generate_empty_activity(message.year)
        message.widget.heatmap.process_data(fake_data)

    @cached_property
    def preview_panel(self) -> Static:
        return self.query_one("#preview", Static)


def main() -> None:
    TimepieceDemo().run()


if __name__ == "__main__":
    main()



from typing import Any

from griffe import Extension, Class

class DefaultCSSFormatter(Extension):
    def on_class_members(self, *, cls: Class, **kwargs: Any) -> None:
        if (attr := cls.attributes.get("DEFAULT_CSS")) is not None:
            attr.value = attr.value.replace('\\n', '\n')[1:-1]



from functools import partial

import pytest

from textual_timepiece._activity_heatmap import HeatmapCursor
from textual_timepiece.activity_heatmap import ActivityHeatmap
from textual_timepiece.activity_heatmap import HeatmapManager


@pytest.fixture
def heatmap_app(create_app):
    return create_app(ActivityHeatmap)


@pytest.fixture
def heatmap_manager_app(create_app):
    return create_app(HeatmapManager)


@pytest.fixture
def heatmap_snap(snap_compare):
    return partial(snap_compare, terminal_size=(165, 21))


@pytest.fixture
def heatmap_data(freeze_time):
    return ActivityHeatmap.generate_empty_activity(freeze_time.year)


@pytest.mark.snapshot
def test_heatmap_hover(heatmap_app, heatmap_snap, heatmap_data, freeze_time):
    async def run_before(pilot):
        heatmap_app.widget.process_data(heatmap_data)
        await pilot.hover(heatmap_app.widget, (17, 1))

    assert heatmap_snap(heatmap_app, run_before=run_before)


@pytest.mark.snapshot
def test_heatmap_hover_month(
    heatmap_app, heatmap_snap, heatmap_data, freeze_time
):
    async def run_before(pilot):
        heatmap_app.widget.process_data(heatmap_data)
        await pilot.hover(heatmap_app.widget, (147, 17))

    assert heatmap_snap(heatmap_app, run_before=run_before)


@pytest.mark.unit
async def test_heatmap_month_nav(heatmap_app, heatmap_data, freeze_time):
    async with heatmap_app.run_test() as pilot:
        heatmap_app.widget.process_data(heatmap_data)
        heatmap_app.widget.focus()
        heatmap_app.widget.cursor = HeatmapCursor(week=0, day=9, month=1)
        assert heatmap_app.widget.cursor.to_date(
            freeze_time.year
        ) == freeze_time.replace(month=1, day=1)

        # NOTE: Additional left for checking date restrictions.
        await pilot.press("left", "right")
        assert heatmap_app.widget.cursor.to_date(
            freeze_time.year
        ) == freeze_time.replace(month=2, day=1)

        await pilot.press("right", "left")
        assert heatmap_app.widget.cursor.to_date(
            freeze_time.year
        ) == freeze_time.replace(month=2, day=1)


@pytest.mark.snapshot
def test_heatmap_manager(
    heatmap_manager_app, heatmap_snap, heatmap_data, freeze_time
):
    async def run_before(pilot):
        heatmap_manager_app.widget.heatmap.process_data(heatmap_data)
        heatmap_manager_app.widget.heatmap.focus()
        await pilot.press("left", "left", "down", "down", "up", "right")

    assert heatmap_snap(heatmap_manager_app, run_before=run_before)


@pytest.mark.unit
async def test_heatmap_navigation(
    heatmap_manager_app, heatmap_snap, heatmap_data, freeze_time
):
    async with heatmap_manager_app.run_test() as pilot:
        heatmap_manager_app.widget.heatmap.process_data(heatmap_data)
        heatmap_manager_app.widget.focus()
        await pilot.press("enter", "tab")
        assert heatmap_manager_app.widget.year == freeze_time.year - 5

        await pilot.press("enter", "tab")
        assert heatmap_manager_app.widget.year == freeze_time.year - 6

        await pilot.press(
            *["backspace"] * 5, "2", "0", "3", "4", "enter", "tab"
        )
        assert heatmap_manager_app.widget.year == 2034

        await pilot.press("enter", *["tab"] * 3)
        assert heatmap_manager_app.widget.year == freeze_time.year

        await pilot.press("enter", "tab")
        assert heatmap_manager_app.widget.year == freeze_time.year + 1

        await pilot.press("enter")
        assert heatmap_manager_app.widget.year == freeze_time.year + 6



from __future__ import annotations

import inspect
from dataclasses import dataclass
from functools import cached_property
from typing import ClassVar
from typing import Iterator
from typing import Self

from rich.console import RenderableType
from rich.console import Text
from textual.app import RenderResult
from textual.dom import DOMNode
from textual.geometry import Offset
from textual.message import Message
from textual.reactive import reactive
from textual.reactive import var
from textual.widget import Widget
from textual.widgets import Button

from textual_timepiece.constants import LOCKED_ICON
from textual_timepiece.constants import TARGET_ICON
from textual_timepiece.constants import UNLOCKED_ICON


class BaseWidget(Widget):
    """Base class with a bunch of utility methods."""

    async def recompose(self) -> None:
        self.clear_cached_properties()
        await super().recompose()

    def clear_cached_properties(self) -> None:
        for prop in self.get_cached_properties():
            if hasattr(self, prop):
                delattr(self, prop)

    @classmethod
    def get_cached_properties(cls) -> Iterator[str]:
        for n, v in inspect.getmembers(cls):
            if isinstance(v, cached_property):
                yield n

    def get_line_offset(self, offset: Offset) -> str:
        x = offset.x - int(self.styles.border_left[0] != "")
        index = 0

        for seg in self.render_line(
            offset.y - self._top_border_offset()
        )._segments:
            # REFACTOR: Look for a public method for this.
            index += len(seg.text)
            if index > x:
                return str(seg.text.strip())

        return ""

    def _top_border_offset(self) -> int:
        return int(self.styles.border_top[0] != "")

    def disable(self, *, disable: bool = True) -> Self:
        self.disabled = disable
        return self

    def disable_messages(  # type: ignore[override]  # NOTE: Easier access to disabling in compose.
        self,
        *messages: type[Message],
    ) -> Self:
        super().disable_messages(*messages)
        return self


class LockButton(Button, BaseWidget):
    DEFAULT_CSS: ClassVar[str] = """
    LockButton {
        background: transparent;
        color: auto;
        min-width: 4;
        max-width: 4;
        height: 3;
        content-align-vertical: middle;
        text-style: bold;
        border: none;

        &:focus {
            background-tint: $primary 50%;
        }
    }
    """

    locked = var[bool](False, init=False)
    icon = reactive[Text](Text(LOCKED_ICON), init=False)

    def __init__(
        self,
        *,
        is_locked: bool = True,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        tooltip: RenderableType | None = None,
        action: str | None = None,
        use_variant: bool = False,
    ) -> None:
        self._use_variant = use_variant
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            tooltip=tooltip,
            action=action,
        )
        self.locked = is_locked

    def compute_icon(self) -> Text:
        return Text(
            LOCKED_ICON if self.locked else UNLOCKED_ICON,
            self.rich_style,
        )

    def render(self) -> RenderResult:
        return self.icon

    def press(self) -> Self:
        self.locked = not self.locked
        return super().press()

    def watch_locked(self) -> None:
        if self._use_variant:
            self.variant = "warning" if self.locked else "success"


class ExpandButton(Button):
    """Button with a special icon."""

    DEFAULT_CSS: ClassVar[str] = """
    ExpandButton {
        background: transparent;
        height: 3;
        min-width: 3;
        max-width: 5;
        border: none;
        padding: 0;
    }
    """

    expanded = var[bool](False, init=False)
    icon = reactive[Text](Text, init=False)

    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        disabled: bool = False,
        tooltip: RenderableType | None = None,
        action: str | None = None,
    ) -> None:
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            tooltip=tooltip,
            action=action,
        )

    def compute_icon(self) -> Text:
        return Text("▲" if self.expanded else "▼", self.rich_style)

    def watch_icon(self, icon: Text) -> None:
        self.label = icon


class TargetButton(Button):
    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        disabled: bool = False,
        tooltip: RenderableType | None = None,
        action: str | None = None,
    ) -> None:
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            tooltip=tooltip,
            action=action,
        )

    def render(self) -> RenderResult:
        return Text(TARGET_ICON, self.rich_style)


@dataclass
class BaseMessage(Message):
    """Generic message that overrides the control method."""

    widget: DOMNode

    @property
    def control(self) -> DOMNode:
        return self.widget



from textual.app import App, ComposeResult
from textual import on
from textual.widgets import Label
from textual_timepiece.pickers import DatePicker, DateSelect
from whenever import Date, days


class DateSelectApp(App[None]):

    def compose(self) -> ComposeResult:
        yield DateSelect(Date.today_in_system_tz(), date_range=days(3))
        yield Label(variant="accent")
    
    @on(DateSelect.DateChanged)
    @on(DateSelect.EndDateChanged)
    def on_date_changed(self, message: DateSelect.DateChanged | DateSelect.EndDateChanged) -> None:
        new_content = f"  {message.widget.date} - {message.widget.end_date}  "
        self.query_one(Label).update(new_content)


if __name__ == "__main__":
    DateSelectApp().run()



from textual.app import App, ComposeResult
from textual_timepiece.pickers import DateRangePicker
from whenever import Date


class DateRangePickerApp(App[None]):
    
    def compose(self) -> ComposeResult:
        yield (dp := DateRangePicker(Date.today_in_system_tz()))
        dp.set_reactive(DateRangePicker.expanded, True)


if __name__ == "__main__":
    DateRangePickerApp().run()



import asyncio

import pytest
from textual.pilot import Pilot
from whenever import Date
from whenever import DateDelta

from textual_timepiece.pickers import DatePicker
from textual_timepiece.pickers import DateSelect


@pytest.fixture
def date_app(create_app):
    return create_app(DatePicker)


@pytest.mark.unit
async def test_picker_dialog(date_app):
    async with date_app.run_test() as pilot:
        date_app.action_focus_next()

        await pilot.press("shift+enter")
        assert date_app.query_one(DatePicker).expanded

        await pilot.press("shift+enter")
        assert not date_app.query_one(DatePicker).expanded


@pytest.mark.snapshot
def test_date_dialog(date_app, snap_compare, freeze_time):
    async def run_before(pilot: Pilot):
        date_app.action_focus_next()
        date_app.widget.query_one("#target-default").press()
        await pilot.press("shift+enter")

    assert snap_compare(date_app, run_before=run_before)


@pytest.mark.snapshot
def test_mini_date_dialog(date_app, snap_compare, freeze_time):
    async def run_before(pilot: Pilot):
        date_app.widget.add_class("mini")
        date_app.action_focus_next()
        date_app.widget.query_one("#target-default").press()
        await pilot.press("shift+enter")

    assert snap_compare(date_app, run_before=run_before)


@pytest.mark.snapshot
def test_date_dialog_range(create_app, snap_compare, freeze_time):
    date_app = create_app(
        DatePicker(freeze_time, date_range=DateDelta(weeks=2))
    )

    async def run_before(pilot: Pilot):
        date_app.action_focus_next()
        await pilot.press("shift+enter")

    assert snap_compare(date_app, run_before=run_before)


@pytest.mark.unit
async def test_target_today(date_app, freeze_time) -> None:
    async with date_app.run_test():
        date_app.widget.query_one("#target-default").press()
        await asyncio.sleep(0.1)
        assert date_app.widget.date == freeze_time


@pytest.mark.unit
async def test_keyboard_dialog_navigaton(date_app, freeze_time) -> None:
    async with date_app.run_test() as pilot:
        date_app.widget.focus().query_one("#target-default").press()
        await pilot.press(
            "shift+enter",
            "right",
            "enter",
            "enter",
            "right",
            "right",
            "enter",
            "down",
            "enter",
            "enter",
            "enter",
        )

        assert date_app.widget.date == Date(2032, 3, 3)


@pytest.mark.unit
async def test_spinbox_features(date_app, freeze_time) -> None:
    async with date_app.run_test() as pilot:
        date_app.widget.query_one("#target-default").focus()

        await pilot.press("enter", "left", "left", "up")
        assert date_app.widget.date == Date(2025, 2, 7)

        await pilot.press("enter", "left", "left", "up")
        assert date_app.widget.date == Date(2025, 3, 7)

        await pilot.press("enter", "left", "left", "down")
        assert date_app.widget.date == Date(2024, 3, 7)


@pytest.mark.unit
async def test_clear_action(date_app) -> None:
    async with date_app.run_test() as pilot:
        date_app.widget.query_one("#target-default").focus()
        await pilot.press("enter", "ctrl+shift+d")
        assert date_app.widget.value is None


@pytest.mark.unit
async def test_date_range(date_app, freeze_time) -> None:
    async with date_app.run_test() as pilot:
        date_app.widget.query_one("#target-default").focus()
        select = date_app.widget.overlay.date_select
        await pilot.press("enter")
        select.date_range = DateDelta(days=5)
        assert date_app.widget.date == freeze_time
        assert select.end_date == Date(2025, 2, 11)


@pytest.mark.unit
async def test_date_dialog_hotkeys(date_app, freeze_time):
    async with date_app.run_test() as pilot:
        date_app.widget.query_one("#target-default").focus()
        select = date_app.widget.overlay.date_select
        select.date_range = DateDelta(days=5)

        # NOTE: Test Default
        await pilot.press("ctrl+t")
        assert date_app.widget.date == freeze_time
        assert select.end_date == Date(2025, 2, 11)

        # NOTE: Test Clear
        await pilot.press("ctrl+shift+d")
        assert date_app.widget.date is None


@pytest.mark.unit
async def test_date_dialog_edge_cases(date_app, freeze_time):
    async with date_app.run_test() as pilot:
        date_app.widget.focus()
        select = date_app.widget.overlay.date_select
        select.post_message(DateSelect.DateChanged(select, Date.MIN))
        await pilot.pause()
        assert date_app.widget.date == Date.MIN

        select.post_message(DateSelect.DateChanged(select, Date.MAX))
        await pilot.pause()
        assert date_app.widget.date == Date.MAX



from textual.app import App, ComposeResult
from textual_timepiece.pickers import DateTimeRangePicker
from textual import on
from whenever import Date, SystemDateTime


class DTPickerRangeApp(App[None]):

    def on_mount(self) -> None:
        self.query_one(DateTimeRangePicker).disable_end()
        self.set_timer(10, self.stop_timer)
        self.notify("Started timer!")

    def compose(self) -> ComposeResult:
        yield DateTimeRangePicker(SystemDateTime.now().local())
    
    def stop_timer(self) -> None:
        dt_range = self.query_one(DateTimeRangePicker).disable_end(disable=False)
        dt_range.end_dt = SystemDateTime.now().local()


if __name__ == "__main__":
    DTPickerRangeApp().run()



from textual.app import App, ComposeResult
from textual_timepiece.pickers import DateTimeDurationPicker
from whenever import LocalDateTime


class DateTimeDurationPickerApp(App[None]):
    
    def compose(self) -> ComposeResult:
        now = LocalDateTime(2025, 2, 10, 12, 13)
        yield (dp := DateTimeDurationPicker(now, now.add(days=1, hours=5, ignore_dst=True)))
        dp.set_reactive(DateTimeDurationPicker.expanded, True)


if __name__ == "__main__":
    DateTimeDurationPickerApp().run()



"""
Helper script to help document all widgets.

This goes through all textual_timepiece widgets and prints the scaffolding for
the tables that are used to document the classvars BINDINGS and
COMPONENT_CLASSES.

Original from https://github.com/Textualize/textual/blob/main/tools/widget_documentation.py
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import textual_timepiece.activity_heatmap
import textual_timepiece.pickers
import textual_timepiece.pickers._base_picker

if TYPE_CHECKING:
    from textual.binding import Binding


# TODO: Gather default CSS and nested bindings.
# TODO: Gather nested bindings


def print_bindings(widget: str, bindings: list[Binding]) -> None:
    """Print a table summarising the bindings.

    The table contains columns for the key(s) that trigger the binding,
    the method that it calls (and tries to link it to the widget itself),
    and the description of the binding.
    """
    if bindings:
        print("BINDINGS")
        print(f'"""All bindings for {widget}\n')
        print("| Key(s) | Description |")
        print("| :- | :- |")

    for binding in bindings:
        print(f"| {binding.key} | {binding.tooltip} |")

    if bindings:
        print('"""')


def print_component_classes(widget: str, classes: set[str]) -> None:
    """Print a table to document these component classes.

    The table contains two columns, one with the component class name and
    another for the description of what the component class is for.
    The second column is always empty.
    """
    if classes:
        print("COMPONENT_CLASSES")
        print(f'"""All component classes for {widget}\n')
        print("| Class | Description |")
        print("| :- | :- |")

    for cls in sorted(classes):
        print(f"| `{cls}` | XXX |")

    if classes:
        print('"""')


def module() -> None:
    """Main entrypoint.

    Iterates over all widgets and prints docs tables.
    """

    modules = [
        textual_timepiece.activity_heatmap,
        textual_timepiece.pickers,
        textual_timepiece.pickers._base_picker,
    ]

    for module in modules:
        print(module.__name__)
        for name in module.__all__:
            w = getattr(module, name)
            bindings: list[Binding] = w.__dict__.get("BINDINGS", [])
            component_classes: set[str] = getattr(
                w, "COMPONENT_CLASSES", set()
            )

            if bindings or component_classes:
                print(name)
                print()
                print_bindings(name, bindings)
                print_component_classes(name, component_classes)
                print()


def main() -> None:
    module()


if __name__ == "__main__":
    main()



from textual.app import App, ComposeResult
from textual_timepiece.pickers import DurationPicker
from whenever import TimeDelta


class DurationPickerApp(App[None]):
    
    def compose(self) -> ComposeResult:
        yield (dp := DurationPicker(TimeDelta()))
        dp.set_reactive(DurationPicker.expanded, True)


if __name__ == "__main__":
    DurationPickerApp().run()



from textual.app import App, ComposeResult
from textual_timepiece.pickers import DatePicker


class DatePickerApp(App[None]):

    def compose(self) -> ComposeResult:
        yield DatePicker()

    def on_date_picker_date_changed(self, message: DatePicker.DateChanged) -> None:
        message.stop()
        if message.date:
            msg = f"Date changed to {message.date.format_common_iso()}."
        else:
            msg = "Date was removed."

        self.notify(msg)


if __name__ == "__main__":
    DatePickerApp().run()



from textual.app import App, ComposeResult
from textual_timepiece.pickers import DateTimeDurationPicker
from whenever import Date, SystemDateTime, weeks


class MiniPickerApp(App[None]):

    def compose(self) -> ComposeResult:
        yield DateTimeDurationPicker(SystemDateTime.now().local(), classes="mini")


if __name__ == "__main__":
    MiniPickerApp().run()



from __future__ import annotations

from contextlib import suppress
from dataclasses import dataclass
from string import digits
from typing import ClassVar
from typing import Final
from typing import Literal
from typing import cast

from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.binding import BindingType
from textual.containers import Grid
from textual.containers import Horizontal
from textual.events import Mount
from textual.reactive import var
from textual.validation import Failure
from textual.validation import ValidationResult
from textual.validation import Validator
from textual.widgets import Button
from textual.widgets import Input
from textual.widgets import Rule
from whenever import SystemDateTime
from whenever import Time
from whenever import TimeDelta
from whenever import hours
from whenever import minutes
from whenever import seconds

from textual_timepiece._extra import BaseMessage
from textual_timepiece._extra import TargetButton
from textual_timepiece._types import Directions
from textual_timepiece._utility import add_time
from textual_timepiece._utility import format_seconds
from textual_timepiece._utility import round_time

from ._base_picker import AbstractInput
from ._base_picker import BaseOverlay
from ._base_picker import BaseOverlayWidget
from ._base_picker import BasePicker


class DurationSelect(BaseOverlayWidget):
    """Duration picker with various buttons in order to set time.

    Params:
        name: Name of the widget.
        id: Unique dom id for the widget
        classes: Any CSS classes that should be added to the widget.
        disabled: Whether to disable the widget.
    """

    @dataclass
    class DurationAdjusted(BaseMessage):
        """Message sent when duration is added or subtracted."""

        widget: DurationSelect
        delta: TimeDelta

    @dataclass
    class DurationRounded(BaseMessage):
        """Sent when one of the headers are clicked in order to round the
        value.
        """

        widget: DurationSelect
        value: int
        """Value used as a rounding factor."""
        scope: Literal["hours", "minutes", "seconds"]
        """Which subunit to round the duration to."""

    DEFAULT_CSS: ClassVar[str] = """
    DurationSelect {
        height: 3;
        layout: horizontal;
        width: 38;

        Grid {
            height: 3;
            grid-size: 2 3;
            grid-gutter: 0;
            grid-rows: 1;

            & > Button:first-of-type {
                column-span: 2;
                text-style: bold;
                background-tint: $background 50%;
            }

            & > Button {
                border: none;
                min-width: 5;
                width: 100%;
                text-style: italic;
                &:hover {
                    border: none;
                }
            }

        }
    }
    """
    """Default CSS for the `DurationSelect` widget."""

    GRID_TEMPLATE: ClassVar[dict[str, tuple[str, ...]]] = {
        "hour-grid": ("Hours", "+1", "+4", "-1", "-4"),
        "minute-grid": ("Minutes", "+15", "+30", "-15", "-30"),
        "second-grid": ("Seconds", "+15", "+30", "-15", "-30"),
    }

    def compose(self) -> ComposeResult:
        for grid, buttons in DurationSelect.GRID_TEMPLATE.items():
            with Grid(id=grid):
                for button in buttons:
                    yield Button(button, classes=grid)

    def _on_button_pressed(self, message: Button.Pressed) -> None:
        message.stop()
        label = str(message.button.label)
        try:
            value = int(label[1:])
            if label.startswith("-"):
                value *= -1
        except ValueError:
            value = None

        if message.button.has_class("hour-grid"):
            if value:
                self.post_message(
                    self.DurationAdjusted(self, TimeDelta(hours=value))
                )
            else:
                self.post_message(self.DurationRounded(self, 21600, "hours"))

        elif message.button.has_class("minute-grid"):
            if value:
                self.post_message(
                    self.DurationAdjusted(self, TimeDelta(minutes=value))
                )
            else:
                self.post_message(self.DurationRounded(self, 3600, "minutes"))
        elif message.button.has_class("second-grid"):
            if value:
                self.post_message(
                    self.DurationAdjusted(self, TimeDelta(seconds=value))
                )
            else:
                self.post_message(self.DurationRounded(self, 60, "seconds"))


class TimeSelect(BaseOverlayWidget):
    """Time selection interface.

    Params:
        name: Name of the widget.
        id: Unique dom id for the widget
        classes: Any CSS classes that should be added to the widget.
        disabled: Whether to disable the widget.
    """

    @dataclass
    class TimeSelected(BaseMessage):
        """Message sent when a value is picked out of the time grid."""

        widget: TimeSelect
        target: Time

    DEFAULT_CSS: ClassVar[str] = """
    TimeSelect {
        layout: grid !important;
        grid-size: 4;
        grid-gutter: 0;
        grid-rows: 1;
        & > Button {
            border: none;
            min-width: 5;
            width: 100%;
            text-style: italic;
            &.dual-hour {
                background: $panel;
            }
            &:focus {
                text-style: bold;
                color: $primary;
            }
            &:hover {
                border: none;
            }
        }
    }
    """
    """Default CSS for the `TimeSelect` widget."""

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding(
            "up",
            "focus_neighbor('up')",
            "Go Up",
            tooltip="Focus the neighbor above.",
            show=False,
        ),
        Binding(
            "right",
            "focus_neighbor('right')",
            "Go Right",
            tooltip="Focus the neighbor to the right.",
            show=False,
        ),
        Binding(
            "down",
            "focus_neighbor('down')",
            "Go Down",
            tooltip="Focus the neighbor below.",
            show=False,
        ),
        Binding(
            "left",
            "focus_neighbor('left')",
            "Go Left",
            tooltip="Focus the neighbor to the left.",
            show=False,
        ),
    ]
    """All bindings for TimeSelect

    | Key(s) | Description |
    | :- | :- |
    | up | Focus the neighbor above. |
    | right | Focus the neighbor to the right. |
    | down | Focus the neighbor below. |
    | left | Focus the neighbor to the left. |
    """

    def compose(self) -> ComposeResult:
        start = Time()
        interval = minutes(30)
        for time in range(48):
            yield Button(
                start.format_common_iso().removesuffix(":00"),
                id=f"time-{time}",
                classes="time icon",
            ).set_class(bool(time % 2), "dual-hour", update=False)
            start = add_time(start, interval)

    def _on_button_pressed(self, message: Button.Pressed) -> None:
        message.stop()
        time = Time.parse_common_iso(f"{message.button.label}:00")
        self.post_message(self.TimeSelected(self, time))

    def action_focus_neighbor(self, direction: Directions) -> None:
        """Focus a nearby member. It will mirror back if going past an edge."""
        if not self.has_focus_within:
            widget = self.query_one("#time-0", Button)
        else:
            # FIX: Subclass a button with a required id.
            focused_id = int(
                cast(str, cast(Button, self.app.focused).id).split("-")[-1]
            )

            row, col = divmod(focused_id, 4)
            if direction == "up":
                if row - 1 >= 0:
                    id = focused_id - 4
                else:
                    id = 43 + (col + 1)

            elif direction == "right":
                if col + 1 <= 3:
                    id = focused_id + 1
                else:
                    id = focused_id - col

            elif direction == "down":
                if row + 1 < 12:
                    id = focused_id + 4
                else:
                    id = col

            else:
                if col - 1 >= 0:
                    id = focused_id - 1
                else:
                    id = focused_id + (3 - col)

            widget = self.query_one(f"#time-{id}", Button)

        self.app.set_focus(widget)


class DurationOverlay(BaseOverlay):
    """Basic dialog widget for editing durations."""

    def compose(self) -> ComposeResult:
        yield DurationSelect()


class TimeOverlay(BaseOverlay):
    """Time dialog which include a time matrix as well."""

    DEFAULT_CSS: ClassVar[str] = """
    TimeOverlay {
        max-width: 40;

        Rule#divider {
            margin: 0;
            padding: 0;
        }
    }
    """

    def compose(self) -> ComposeResult:
        yield DurationSelect()
        yield Rule.horizontal(id="divider")
        yield TimeSelect()


class DurationInput(AbstractInput[TimeDelta]):
    """Duration input for time deltas."""

    @dataclass
    class DurationChanged(BaseMessage):
        """Message sent when the duration changes through input or spinbox."""

        widget: DurationInput
        duration: TimeDelta | None

    ALIAS = "duration"
    MIN: Final[TimeDelta] = TimeDelta()
    MAX: Final[TimeDelta] = TimeDelta(hours=99, minutes=59, seconds=59)

    PATTERN = "99:99:99"

    duration = var[TimeDelta | None](None, init=False)
    """Current duration in a `TimeDelta` or None if empty.

    This value is capped at 99 hours, 59 minutes and 59 seconds.
    """

    def _validate_duration(
        self,
        duration: TimeDelta | None,
    ) -> TimeDelta | None:
        if duration is None:
            return None

        return max(self.MIN, min(self.MAX, duration))

    def _watch_duration(self, duration: TimeDelta | None) -> None:
        with self.prevent(Input.Changed), suppress(ValueError):
            if isinstance(duration, TimeDelta):
                self.value = format_seconds(int(duration.in_seconds()))
            else:
                self.value = ""

        self.post_message(self.DurationChanged(self, self.duration))

    def _watch_value(self, value: str) -> None:
        if dur := self.convert():
            self.duration = dur
        super()._watch_value(value)

    def convert(self) -> TimeDelta | None:
        try:
            hours, minutes, seconds = tuple(map(int, self.value.split(":")))
        except ValueError:
            return None
        if hours + minutes + seconds == 0:
            return TimeDelta()
        return TimeDelta(seconds=seconds, minutes=minutes, hours=hours)

    def action_adjust_time(self, offset: int) -> None:
        """Adjust time with an offset depending on the text cursor position."""
        if self.duration is None:
            self.duration = TimeDelta()
        elif 0 <= self.cursor_position < 2:
            self.duration += hours(offset)
        elif 3 <= self.cursor_position < 5:
            self.duration += minutes(offset)
        elif 6 <= self.cursor_position:
            self.duration += seconds(offset)


# TODO: Find a better way to show larger durations.


class DurationPicker(BasePicker[DurationInput, TimeDelta, DurationOverlay]):
    """Picker widget for picking durations.

    Duration is limited 99 hours, 59 minutes and 59 seconds.

    Params:
        value: Initial duration value for the widget.
        name: Name for the widget.
        id: DOM identifier for the widget.
        classes: CSS classes for the widget
        disabled: Whether to disable the widget.
        tooltip: Tooltip to show on hover.
    """

    @dataclass
    class DurationChanged(BaseMessage):
        """Message sent when the duration changes."""

        widget: DurationPicker
        duration: TimeDelta | None

    INPUT = DurationInput
    ALIAS = "duration"

    duration = var[TimeDelta | None](None, init=False)
    """Current duration. Bound to all the child widgets."""

    def compose(self) -> ComposeResult:
        with Horizontal(id="input-control"):
            yield (
                DurationInput(id="data-input").data_bind(
                    duration=DurationPicker.duration
                )
            )
            yield TargetButton(
                id="target-default",
                tooltip="Set the duration to zero.",
            )
            yield self._compose_expand_button()

        yield DurationOverlay().data_bind(show=DurationPicker.expanded)

    def _on_mount(self, event: Mount) -> None:
        self.query_exactly_one("#target-default", Button).disabled = (
            self.duration is None or self.duration.in_seconds() == 0
        )

    def _watch_duration(self, delta: TimeDelta) -> None:
        self.query_exactly_one("#target-default", Button).disabled = (
            delta is None or delta.in_seconds() == 0
        )
        self.post_message(self.DurationChanged(self, delta))

    @on(DurationSelect.DurationRounded)
    def _round_duration(self, message: DurationSelect.DurationRounded) -> None:
        message.stop()
        if self.duration is None:
            return
        seconds = (
            round(self.duration.in_seconds() / message.value) * message.value
        )
        self.duration = TimeDelta(seconds=seconds)

    @on(DurationSelect.DurationAdjusted)
    def _adjust_duration(
        self,
        message: DurationSelect.DurationAdjusted,
    ) -> None:
        message.stop()
        if message.delta is None:
            self.duration = None
        elif self.duration is None:
            self.duration = message.delta
        else:
            self.duration += message.delta

    @on(DurationInput.DurationChanged)
    def _set_duration(self, message: DurationInput.DurationChanged) -> None:
        message.stop()
        with message.control.prevent(DurationInput.DurationChanged):
            self.duration = message.duration

    def to_default(self) -> None:
        """Reset the duration to 00:00:00."""
        self.duration = TimeDelta()


class TimeValidator(Validator):
    def validate(self, value: str) -> ValidationResult:
        try:
            Time.parse_common_iso(value.strip())
        except ValueError:
            return ValidationResult(
                [
                    Failure(
                        self,
                        "Parsing Failure!",
                        "Value is not adhering to ISO time!",
                    )
                ]
            )

        return self.success()


class TimeInput(AbstractInput[Time]):
    """Time input for a HH:MM:SS format"""

    @dataclass
    class TimeChanged(BaseMessage):
        """Message sent when the time is updated."""

        widget: TimeInput
        new_time: Time | None

    PATTERN = "00:00:00"
    ALIAS = "time"

    time = var[Time | None](None, init=False)
    """Currently set time or none if its empty."""

    def _watch_time(self, time: Time | None) -> None:
        with self.prevent(Input.Changed), suppress(ValueError):
            if time:
                self.value = time.format_common_iso()
            else:
                self.value = ""

        self.post_message(self.TimeChanged(self, self.time))

    def _watch_value(self, value: str) -> None:
        if (ts := self.convert()) is not None and ts != self.time:
            self.time = ts
        super()._watch_value(value)

    def convert(self) -> Time | None:
        try:
            return Time.parse_common_iso(self.value)
        except ValueError:
            return None

    def insert_text_at_cursor(self, text: str) -> None:
        if self.cursor_position > 7:
            return

        if text not in digits:
            self.cursor_position += 1
            return

        if self.cursor_position == 0:
            self.value = text + self.value[1:]
        elif self.cursor_position == 7:
            self.value = self.value[:-1] + text
            return
        elif self.cursor_position in {4, 1} or (
            self.value[self.cursor_position] != ":" and text in "543210"
        ):
            self.value = (
                self.value[: self.cursor_position]
                + text
                + self.value[self.cursor_position + 1 :]
            )

        self.cursor_position += 1

    def action_delete_right(self) -> None:
        return

    def action_delete_left(self) -> None:
        if self.cursor_position < 0:
            return

        elif self.cursor_position == 8:
            self.value = self.value[:-1] + "0"

        elif self.cursor_position not in {2, 5}:
            self.value = (
                self.value[: self.cursor_position]
                + "0"
                + self.value[self.cursor_position + 1 :]
            )

        if self.cursor_position > 0:
            self.cursor_position -= 1

    def action_adjust_time(self, offset: int) -> None:
        """Adjust time with an offset depending on the text cursor position."""
        if 0 <= self.cursor_position < 2:
            self.time = add_time(self.time or Time(), hours(offset))
        elif 3 <= self.cursor_position < 5:
            self.time = add_time(self.time or Time(), minutes(offset))
        elif 6 <= self.cursor_position:
            self.time = add_time(self.time or Time(), seconds(offset))


class TimePicker(BasePicker[TimeInput, Time, TimeOverlay]):
    """Time picker for a 24 hour clock.

    Params:
        value: Initial time for the widget.
        name: Name for the widget.
        id: DOM identifier for the widget.
        classes: CSS classes for the widget
        disabled: Whether to disable the widget.
        tooltip: Tooltip to show on hover.
    """

    @dataclass
    class TimeChanged(BaseMessage):
        """Sent when the time is changed with the overlay or other means."""

        widget: TimePicker
        new_time: Time

    INPUT = TimeInput
    ALIAS = "time"

    time = var[Time | None](None, init=False)
    """Currently set time that is bound to the subwidgets. None if empty."""

    def compose(self) -> ComposeResult:
        with Horizontal(id="input-control"):
            yield TimeInput(id="data-input").data_bind(time=TimePicker.time)
            yield TargetButton(id="target-default", tooltip="Set time to now.")
            yield self._compose_expand_button()

        yield TimeOverlay().data_bind(show=TimePicker.expanded)

    @on(DurationSelect.DurationRounded)
    def _round_duration(self, message: DurationSelect.DurationRounded) -> None:
        message.stop()
        if self.time is None:
            return
        self.time = round_time(self.time, message.value)

    @on(DurationSelect.DurationAdjusted)
    def _adjust_duration(
        self, message: DurationSelect.DurationAdjusted
    ) -> None:
        message.stop()
        self.time = add_time(self.time or Time(), message.delta)

    @on(TimeSelect.TimeSelected)
    def _select_time(self, message: TimeSelect.TimeSelected) -> None:
        message.stop()
        self.time = message.target

    @on(TimeInput.TimeChanged)
    def _change_time(self, message: TimeInput.TimeChanged) -> None:
        message.stop()
        with message.control.prevent(TimeInput.TimeChanged):
            self.time = message.new_time

    def to_default(self) -> None:
        """Reset time to the local current time."""
        self.time = SystemDateTime.now().time()



from textual.app import App, ComposeResult
from textual_timepiece.pickers import DateTimeRangePicker
from whenever import Date, SystemDateTime


class DateTimeRangePickerApp(App[None]):
    
    def compose(self) -> ComposeResult:
        now = SystemDateTime.now().local()
        yield (dp := DateTimeRangePicker(now, now.add(hours=14, ignore_dst=True)))
        dp.set_reactive(DateTimeRangePicker.expanded, True)


if __name__ == "__main__":
    DateTimeRangePickerApp().run()



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import ClassVar
from typing import Self
from typing import cast

from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.binding import BindingType
from textual.containers import Horizontal
from textual.containers import Vertical
from textual.reactive import var
from textual.widgets import Button
from textual.widgets import Static
from whenever import Date
from whenever import DateDelta
from whenever import LocalDateTime
from whenever import SystemDateTime
from whenever import Time
from whenever import TimeDelta

from textual_timepiece._extra import BaseMessage
from textual_timepiece._extra import LockButton
from textual_timepiece._extra import TargetButton
from textual_timepiece._utility import round_time

from ._base_picker import AbstractPicker
from ._base_picker import BaseOverlay
from ._date_picker import DateInput
from ._date_picker import DateSelect
from ._date_picker import EndDateSelect
from ._datetime_picker import DateTimeInput
from ._time_picker import DurationInput
from ._time_picker import DurationSelect

# TODO: Set a maximum and minimum range that is required.
# TODO: Limit or validate to min/max dates


class DateRangeOverlay(BaseOverlay):
    """Simple date range dialog with to date selects combined."""

    DEFAULT_CSS: ClassVar[str] = """
    DateRangeOverlay {
        layout: horizontal;

        Static#spacer {
            width: 1;
            height: 100%;
            hatch: ">" $primary 50%;
        }
    }
    """
    start = var[Date | None](None, init=False)
    stop = var[Date | None](None, init=False)

    def compose(self) -> ComposeResult:
        yield DateSelect(is_range=True, id="start-date-select").data_bind(
            date=DateRangeOverlay.start, end_date=DateRangeOverlay.stop
        )
        yield Static(id="spacer")
        yield EndDateSelect(id="end-date-select").data_bind(
            date=DateRangeOverlay.start, end_date=DateRangeOverlay.stop
        )


class DateRangePicker(AbstractPicker[DateRangeOverlay]):
    """Date range picker for picking inclusive date ranges.

    Params:
        start: Initial start date for the picker.
        end: Initial end date for the picker.
        name: Name for the widget.
        id: DOM identifier for the widget.
        classes: CSS classes for the widget
        date_range: Date range to restrict the date to. If provided the picker
            lock will be permanently on for the widgets lifetime or when
            re-enabled programmatically.
        disabled: Whether to disable the widget
        tooltip: Tooltip to show on hover.

    Examples:
        ```python
            def compose(self) -> ComposeResult:
                yield DateRangePicker(Date(2025, 2, 1), Date(2025, 3, 1))
        ```

        ```python
            def compose(self) -> ComposeResult:
                yield DateRangePicker(Date.today_in_system_tz()).disable_end()

            def action_stop(self) -> None:
                pick = self.query_one(DateRangePicker)
                pick.disable_end(disable=False)
                pick.end_date = Date.today_in_system_tz()
        ```
    """

    @dataclass
    class DateRangeChanged(BaseMessage):
        """Message sent when the date range has changed."""

        widget: DateRangePicker
        start: Date | None
        end: Date | None

    BINDING_GROUP_TITLE = "Date Range Picker"

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding(
            "ctrl+shift+d",
            "clear",
            "Clear Dates",
            tooltip="Clear both the start and end date.",
        ),
        Binding(
            "ctrl+t",
            "target_default_start",
            "Start To Today",
            tooltip="Set the start date to todays date.",
        ),
        Binding(
            "alt+ctrl+t",
            "target_default_end",
            "End To Today",
            tooltip="Set the end date to today or the start date.",
        ),
    ]
    """All bindings for `DateTimeRangePicker`.

    | Key(s) | Description |
    | :- | :- |
    | ctrl+shift+d | Clear end and start datetime. |
    | ctrl+t | Set the start date to todays date. |
    | alt+ctrl+t | Set the end date to today or the start date. |
    """

    start_date = var[Date | None](None, init=False)
    """Picker start date. Bound to sub widgets."""
    end_date = var[Date | None](None, init=False)
    """Picker end date. Bound to sub widgets."""

    def __init__(
        self,
        start: Date | None = None,
        end: Date | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        date_range: DateDelta | None = None,
        disabled: bool = False,
        tooltip: str | None = None,
    ) -> None:
        super().__init__(name, id, classes, disabled=disabled, tooltip=tooltip)
        self.set_reactive(DateRangePicker.start_date, start)
        self.set_reactive(DateRangePicker.end_date, end)
        self._date_range = date_range

    def compose(self) -> ComposeResult:
        with Horizontal(id="input-control"):
            yield DateInput(id="start-date-input").data_bind(
                date=DateRangePicker.start_date,
            )

            yield TargetButton(
                id="target-default-start",
                tooltip="Set the start date to today.",
            )
            yield LockButton(
                is_locked=self._date_range is not None,
                id="lock-button",
                tooltip="Lock the range inbetween the dates.",
                disabled=self._date_range is not None,
            )

            yield DateInput(id="stop-date-input").data_bind(
                date=DateRangePicker.end_date,
            )
            yield TargetButton(
                id="target-default-end",
                tooltip="Set the end date to today or the start date.",
            )
            yield self._compose_expand_button()

        yield DateRangeOverlay().data_bind(
            show=DateRangePicker.expanded,
            start=DateRangePicker.start_date,
            stop=DateRangePicker.end_date,
        )

    def _watch_start_date(self, date: Date | None) -> None:
        if date and self._date_range:
            with self.prevent(self.DateRangeChanged):
                self.end_date = date + self._date_range
        self.post_message(self.DateRangeChanged(self, date, self.end_date))

    def _watch_end_date(self, date: Date | None) -> None:
        if date and self._date_range:
            with self.prevent(self.DateRangeChanged):
                self.start_date = date - self._date_range
        self.post_message(self.DateRangeChanged(self, self.start_date, date))

    @on(DateSelect.DateChanged)
    @on(DateSelect.EndDateChanged)
    def _dialog_date_changed(
        self,
        message: DateSelect.DateChanged | DateSelect.EndDateChanged,
    ) -> None:
        """Handles changes in dates including, keeping dates the same span."""
        message.stop()
        if isinstance(message, DateSelect.DateChanged):
            self.start_date = message.date
        else:
            self.end_date = message.date

    @on(DateInput.DateChanged, "#start-date-input")
    @on(DateInput.DateChanged, "#stop-date-input")
    def _date_input_change(self, message: DateInput.DateChanged) -> None:
        message.stop()
        with message.control.prevent(DateInput.DateChanged):
            if message.control.id == "start-date-input":
                self.start_date = message.date
            else:
                self.end_date = message.date

    @on(Button.Pressed, "#target-default-start")
    def _action_target_default_start(
        self,
        message: Button.Pressed | None = None,
    ) -> None:
        if message:
            message.stop()
        new_date = Date.today_in_system_tz()
        if not self.end_date or new_date <= self.end_date:
            self.start_date = new_date
        else:
            self.start_date = self.end_date

    @on(Button.Pressed, "#target-default-end")
    def _action_target_default_end(
        self,
        message: Button.Pressed | None = None,
    ) -> None:
        if message:
            message.stop()
        new_date = Date.today_in_system_tz()
        if not self.start_date or (new_date) >= self.start_date:
            self.end_date = new_date
        else:
            self.end_date = self.start_date

    @on(Button.Pressed, "#lock-button")
    def _lock_delta(self, message: Button.Pressed) -> None:
        message.stop()

        if (
            self.end_date
            and self.start_date
            and cast(LockButton, message.control).locked
        ):
            self._date_range = self.end_date - self.start_date
        else:
            self._date_range = None
            cast(LockButton, message.control).locked = False

    def action_clear(self) -> None:
        """Clear the start and end dates."""
        self.start_date = None
        self.end_date = None

    def disable_start(self, *, disable: bool = True) -> Self:
        """Utility method to disable start input widgets."""
        self.start_input.disabled = disable
        self.overlay.query_one(
            "#start-date-select", DateSelect
        ).disabled = disable
        self.query_one("#target-default-start", Button).disabled = disable
        return self

    def disable_end(self, *, disable: bool = True) -> Self:
        """Utility method to disable end input widgets."""
        self.end_input.disabled = disable
        self.overlay.query_one(
            "#end-date-select", EndDateSelect
        ).disabled = disable
        self.query_one("#target-default-end", Button).disabled = disable
        return self

    @cached_property
    def start_input(self) -> DateInput:
        return self.query_exactly_one("#start-date-input", DateInput)

    @cached_property
    def end_input(self) -> DateInput:
        return self.query_exactly_one("#stop-date-input", DateInput)

    @cached_property
    def lock_button(self) -> LockButton:
        return cast(LockButton, self.query_exactly_one(LockButton))


class DateTimeRangeOverlay(BaseOverlay):
    DEFAULT_CSS: ClassVar[str] = """
    DateTimeRangeOverlay {
        layout: horizontal !important;
        width: auto;
        Vertical {
            width: auto;
            height: auto;
        }
        Static {
            width: 1;
            height: 100%;
            hatch: "⟩" $primary 50%;
        }
    }
    """
    """Default CSS for the `DateTimeRangeOverlay`."""

    start = var[Date | None](None, init=False)
    stop = var[Date | None](None, init=False)

    def compose(self) -> ComposeResult:
        with Vertical(id="start-column"):
            yield DurationSelect(id="start-time-select")
            yield DateSelect(is_range=True, id="start-date-select").data_bind(
                date=DateTimeRangeOverlay.start,
                end_date=DateTimeRangeOverlay.stop,
            )
        yield Static(id="spacer", expand=True)
        with Vertical(id="end-column"):
            yield DurationSelect(id="end-time-select")
            yield EndDateSelect(id="end-date-select").data_bind(
                date=DateTimeRangeOverlay.start,
                end_date=DateTimeRangeOverlay.stop,
            )


class DateTimeRangePicker(AbstractPicker[DateTimeRangeOverlay]):
    """Datetime range picker with two datetime inputs.

    Params:
        start: Initial start datetime for the picker.
        end: Initial end datetime for the picker.
        name: Name for the widget.
        id: DOM identifier for the widget.
        classes: CSS classes for the widget
        time_range: Time range to restrict the datetimes to. If provided the
            picker lock will be permanently on for the widgets lifetime or
            re-enabled programmatically.
        disabled: Whether to disable the widget
        tooltip: Tooltip to show on hover.

    Examples:
        ```python
            def compose(self) -> ComposeResult:
                now = SystemDateTime.now().local()
                yield DateTimeRangePicker(now, time_range=TimeDelta(hours=5))
        ```

        ```python
            def compose(self) -> ComposeResult:
                yield DateTimeRangePicker(SystemDateTime.now().local())

            def action_stop(self) -> None:
                pick = self.query_one(DateTimeRangePicker)
                pick.end_dt = SystemDateTime.now().local()
        ```
    """

    @dataclass
    class DTRangeChanged(BaseMessage):
        """Message sent when the datetime range has changed."""

        widget: DateTimeRangePicker
        start: LocalDateTime | None
        end: LocalDateTime | None

    BINDING_GROUP_TITLE = "Datetime Range Picker"

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding(
            "ctrl+shift+d",
            "clear",
            "Clear",
            tooltip="Clear end and start datetime.",
        ),
        Binding(
            "ctrl+t",
            "target_default_start",
            "Start To Today",
            tooltip="Set the start datetime to now.",
        ),
        Binding(
            "alt+ctrl+t",
            "target_default_end",
            "End To Today",
            tooltip="Set the end datetime to now or the start datetime.",
        ),
    ]
    """All bindings for `DateTimeRangePicker`.

    | Key(s) | Description |
    | :- | :- |
    | ctrl+shift+d | Clear end and start datetime. |
    | ctrl+t | Set the start datetime to now. |
    | alt+ctrl+t | Set the end datetime to now or the start datetime. |
    """

    start_dt = var[LocalDateTime | None](None, init=False)
    """Picker start datetime. Bound to all the parent widgets."""
    end_dt = var[LocalDateTime | None](None, init=False)
    """Picker end datetime. Bound to all the parent widgets."""

    start_date = var[Date | None](None, init=False)
    """Start date dynamically computed depending on start datetime."""
    end_date = var[Date | None](None, init=False)
    """End date dynamically computed depending on the end datetime."""

    def __init__(
        self,
        start: LocalDateTime | None = None,
        end: LocalDateTime | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        *,
        time_range: TimeDelta | None = None,
        disabled: bool = False,
        tooltip: str | None = None,
    ) -> None:
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            tooltip=tooltip,
        )
        self.set_reactive(DateTimeRangePicker.start_dt, start)
        self.set_reactive(DateTimeRangePicker.end_dt, end)

        self._time_range = time_range

    def compose(self) -> ComposeResult:
        with Horizontal(id="input-control"):
            yield DateTimeInput(id="start-dt-input").data_bind(
                datetime=DateTimeRangePicker.start_dt,
            )

            yield TargetButton(
                id="target-default-start",
                tooltip="Set the start time to now.",
            )
            yield LockButton(
                is_locked=self._time_range is not None,
                id="lock-button",
                tooltip="Lock the time range.",
                disabled=self._time_range is not None,
            )

            yield DateTimeInput(id="end-dt-input").data_bind(
                datetime=DateTimeRangePicker.end_dt,
            )
            yield TargetButton(
                id="target-default-end",
                tooltip="Set the end time to now or the start time.",
            )
            yield self._compose_expand_button()

        yield DateTimeRangeOverlay().data_bind(
            show=DateTimeDurationPicker.expanded,
            start=DateTimeDurationPicker.start_date,
            stop=DateTimeDurationPicker.end_date,
        )

    def _compute_start_date(self) -> Date | None:
        if self.start_dt is None:
            return None
        return self.start_dt.date()

    def _compute_end_date(self) -> Date | None:
        if self.end_dt is None:
            return None

        return self.end_dt.date()

    def _watch_start_dt(self, new: LocalDateTime | None) -> None:
        if new and self._time_range:
            with self.prevent(self.DTRangeChanged):
                self.end_dt = new.add(
                    seconds=self._time_range.in_seconds(),
                    ignore_dst=True,
                )
        self.post_message(self.DTRangeChanged(self, new, self.end_dt))

    def _watch_end_dt(self, new: LocalDateTime | None) -> None:
        if new and self._time_range:
            with self.prevent(self.DTRangeChanged):
                self.start_dt = new.subtract(
                    seconds=self._time_range.in_seconds(),
                    ignore_dst=True,
                )
        self.post_message(self.DTRangeChanged(self, self.start_dt, new))

    @on(Button.Pressed, "#lock-button")
    def _lock_delta(self, message: Button.Pressed) -> None:
        message.stop()

        if (
            cast(LockButton, message.control).locked
            and self.end_dt
            and self.start_dt
        ):
            self._time_range = self.end_dt.difference(
                self.start_dt, ignore_dst=True
            )
        else:
            self._time_range = None
            cast(LockButton, message.control).locked = False

    @on(DateSelect.DateChanged)
    @on(DateSelect.EndDateChanged)
    def _dialog_date_changed(self, message: DateSelect.DateChanged) -> None:
        message.stop()
        if isinstance(message, DateSelect.DateChanged):
            self.adjust_start_date(message.date)
        else:
            self.adjust_end_date(message.date)

    def adjust_start_date(self, date: Date | None) -> None:
        """Set or clear the current start date depending on the input."""
        if self.start_dt and date:
            self.start_dt = self.start_dt.replace_date(date)
        elif date:
            self.start_dt = date.at(Time())
        else:
            self.start_dt = date

    def adjust_end_date(self, date: Date | None) -> None:
        """Set or clear the current end date depending on the input."""
        if self.end_dt and date:
            self.end_dt = self.end_dt.replace_date(date)
        elif date:
            self.end_dt = date.at(Time())
        else:
            self.end_dt = date

    def action_clear(self) -> None:
        """Clear the start and end datetimes."""
        self.start_dt = None
        self.end_dt = None

    @on(DurationSelect.DurationRounded)
    def _round_duration(self, message: DurationSelect.DurationRounded) -> None:
        message.stop()
        if message.widget.id == "start-time-select":
            if self.start_dt is None:
                return
            time = round_time(self.start_dt.time(), message.value)
            self.start_dt = self.start_dt.replace_time(time)

        elif self.end_dt:
            time = round_time(self.end_dt.time(), message.value)
            self.end_dt = self.end_dt.replace_time(time)

    @on(DurationSelect.DurationAdjusted)
    def _adjust_duration(
        self, message: DurationSelect.DurationAdjusted
    ) -> None:
        message.stop()
        if message.widget.id == "start-time-select":
            if self.start_dt is None:
                return
            self.start_dt = self.start_dt.add(message.delta, ignore_dst=True)
        elif self.end_dt:
            self.end_dt = self.end_dt.add(message.delta, ignore_dst=True)
        elif self.start_dt:
            self.end_dt = self.start_dt.add(message.delta, ignore_dst=True)

    @on(DateTimeInput.DateTimeChanged, "#start-dt-input")
    def _start_dt_input_changed(
        self,
        message: DateTimeInput.DateTimeChanged,
    ) -> None:
        message.stop()
        with message.control.prevent(DateTimeInput.DateTimeChanged):
            self.start_dt = message.datetime

    @on(DateTimeInput.DateTimeChanged, "#end-dt-input")
    def _end_dt_input_changed(
        self, message: DateTimeInput.DateTimeChanged
    ) -> None:
        message.stop()
        with message.control.prevent(DateTimeInput.DateTimeChanged):
            self.end_dt = message.datetime

    def disable_start(self, *, disable: bool = True) -> Self:
        """Utility method to disable start input widgets."""
        self.start_input.disabled = disable
        self.overlay.query_one("#start-column", Vertical).disabled = disable
        self.query_one("#target-default-start", Button).disabled = disable
        return self

    def disable_end(self, *, disable: bool = True) -> Self:
        """Utility method to disable end input widgets."""
        self.end_input.disabled = disable
        self.overlay.query_one("#end-column", Vertical).disabled = disable
        self.query_one("#target-default-end", Button).disabled = disable
        return self

    @on(Button.Pressed, "#target-default-start")
    def _action_target_default_start(
        self,
        message: Button.Pressed | None = None,
    ) -> None:
        if message:
            message.stop()
        self.start_dt = SystemDateTime.now().local()

    @on(Button.Pressed, "#target-default-end")
    def _action_target_default_end(
        self,
        message: Button.Pressed | None = None,
    ) -> None:
        if message:
            message.stop()
        now = SystemDateTime.now().local()
        if not self.start_dt or now >= self.start_dt:
            self.end_dt = now
        else:
            self.end_dt = self.start_dt

    @cached_property
    def start_input(self) -> DateTimeInput:
        return self.query_exactly_one("#start-dt-input", DateTimeInput)

    @cached_property
    def end_input(self) -> DateTimeInput:
        return self.query_exactly_one("#end-dt-input", DateTimeInput)

    @cached_property
    def lock_button(self) -> LockButton:
        return cast(LockButton, self.query_exactly_one(LockButton))


class DateTimeDurationPicker(DateTimeRangePicker):
    """Datetime range with a duration input in the middle.

    Duration display up to 99:99:99. Use the DateTimeRangePicker picker if a
    longer duration is required.

    Params:
        start: Initial start datetime for the picker.
        end: Initial end datetime for the picker.
        name: Name for the widget.
        id: DOM identifier for the widget.
        classes: CSS classes for the widget
        time_range: Time range to restrict the datetimes to. If provided the
            picker lock will be permanently on for the widgets lifetime or
            re-enabled programmatically.
        disabled: Whether to disable the widget
        tooltip: Tooltip to show on hover.
    """

    duration = var[TimeDelta | None](None, init=False)
    """Duration between start and end datetimes. Computed dynamically."""

    async def _on_mount(self) -> None:  # type: ignore[override] # NOTE: Need to mount extra widget
        """Overrides the compose method in order to a duration input."""
        await self.query_exactly_one("#input-control", Horizontal).mount(
            DurationInput(self.duration, id="duration-input").data_bind(
                DateTimeDurationPicker.duration
            ),
            after=1,
        )

    def _compute_duration(self) -> TimeDelta:
        if self.start_dt is None or self.end_dt is None:
            return TimeDelta()
        return self.end_dt.difference(self.start_dt, ignore_dst=True)

    @on(DurationInput.DurationChanged)
    def _new_duration(self, message: DurationInput.DurationChanged) -> None:
        message.stop()
        with message.control.prevent(DurationInput.DurationChanged):
            if message.duration is None:
                self.end_dt = None
            elif self.start_dt:
                self.end_dt = self.start_dt.add(
                    seconds=message.duration.in_seconds(),
                    ignore_dst=True,
                )
            elif self.end_dt:
                self.start_dt = self.end_dt.subtract(
                    seconds=message.duration.in_seconds(),
                    ignore_dst=True,
                )

    @on(Button.Pressed, "#lock-button")
    def _lock_duration(self, message: Button.Pressed) -> None:
        message.stop()
        if self.start_date:
            self.duration_input.disabled = cast(
                LockButton, message.button
            ).locked

    @cached_property
    def duration_input(self) -> DurationInput:
        return cast(DurationInput, self.query_exactly_one(DurationInput))



"""Utilities for parsing and managing time related objects."""

from __future__ import annotations

import enum
import math
from calendar import Calendar
from calendar import month_name
from collections.abc import Sequence
from typing import Any
from typing import Iterator
from typing import Protocol
from typing import TypeAlias
from typing import TypeVar

from whenever import Date
from whenever import DateDelta
from whenever import Time
from whenever import TimeDelta


def breakdown_seconds(total_seconds: int) -> tuple[int, int, int]:
    hours, remainder = divmod(total_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)

    return hours, minutes, seconds


def format_seconds(total_seconds: int, *, include_seconds: bool = True) -> str:
    hours, minutes, seconds = breakdown_seconds(total_seconds)
    result = f"{hours:0>2}:{minutes:0>2}"
    if include_seconds:
        result += f":{seconds:0>2}"
    return result


class DateTime(Protocol):
    @property
    def hour(self) -> int: ...
    @property
    def minute(self) -> int: ...
    @property
    def second(self) -> int: ...


def time_to_seconds(ts: DateTime) -> int:
    return (ts.hour * 3600) + (ts.minute * 60) + (ts.second)


class DateScope(enum.Enum):
    MONTH = enum.auto()
    YEAR = enum.auto()
    DECADE = enum.auto()
    CENTURY = enum.auto()


Scope: TypeAlias = list[list[int]] | list[list[str]]


def _get_year_scope() -> list[list[str]]:
    year_scope: list[list[str]] = []
    month_names = list(month_name)[1:]
    for i in range(4):
        year_scope.append([])
        year_scope[i].extend(month_names[i * 3 : i * 3 + 3])
    return year_scope


def _get_decade_scope(period: Date) -> list[list[int]]:
    data: list[list[int]] = []
    year = math.floor(period.year / 10) * 10
    for i, y in enumerate(range(year, year + 10)):
        if i % 3 == 0:
            data.append([])
        data[i // 3].append(y)
    return data


def _get_century_scopee(period: Date) -> list[list[int]]:
    data: list[list[int]] = []
    century = math.floor(period.year / 100) * 100
    for i, year in enumerate(range(century, century + 100, 10)):
        if i % 3 == 0:
            data.append([])
        data[i // 3].append(year)
    return data


def get_scope(scope: DateScope, period: Date) -> Scope:
    """Creates a two dimensional array depending on supplied scope.

    Args:
        scope: The scope of time the data should encompass.
        period: Period to filter for the data.

    Raises:
        NotImplementedError: If the wrong date scope is supplied.

    Returns:
        The range of times in a two dimensional array.
    """

    if scope == DateScope.MONTH:
        return Calendar().monthdayscalendar(period.year, period.month)
    if scope == DateScope.YEAR:
        return _get_year_scope()
    if scope == DateScope.DECADE:
        return _get_decade_scope(period)
    if scope == DateScope.CENTURY:
        return _get_century_scopee(period)

    raise NotImplementedError(f"{scope.name} scope is not implemented!")


def add_time(time: DateTime, delta: TimeDelta) -> Time:
    value = time_to_seconds(time) + int(delta.in_seconds())
    if value <= 0:
        return Time()
    return Time.parse_common_iso(format_seconds(min(value, 86399)))


def round_time(time: DateTime, multiple: int) -> Time:
    seconds = round((time_to_seconds(time) / multiple) * multiple)
    return Time.parse_common_iso(format_seconds(seconds))


def iterate_timespan(
    start: Date,
    increment: DateDelta,
    total: int,
) -> Iterator[Date]:
    for _ in range(total):
        yield start
        try:
            start += increment
        except ValueError:
            return


def _nomalize_value(
    value: float | None, min_val: float, denom: float
) -> float | None:
    if value is None:
        return None
    try:
        return (value - min_val) / denom
    except ZeroDivisionError:
        return 0


def normalize_values(values: Sequence[float | None]) -> list[float | None]:
    """Normalizes an sequence of values to the range of 0 to 1."""
    min_val, max_val = (
        min(v for v in values if v is not None),
        max(v for v in values if v is not None),
    )
    denom = max_val - min_val
    return [
        _nomalize_value(values[i], min_val, denom) for i in range(len(values))
    ]


T = TypeVar("T")


def flat_to_shape(
    values: Sequence[T],
    shape: list[list[Any]],
) -> list[list[T]]:
    """Transform a flat iterable back into a two dimensional one."""
    total = 0
    for row in range(len(shape)):
        for i in range(len(shape[row])):
            shape[row][i] = values[total]
            total += 1

    return shape



from textual.app import App, ComposeResult
from textual import on
from textual_timepiece.pickers import DurationSelect


class DurationSelectApp(App[None]):

    def compose(self) -> ComposeResult:
        yield DurationSelect()
    

if __name__ == "__main__":
    DurationSelectApp().run()



import nox

nox.options.default_venv_backend = "uv|virtualenv"


@nox.session(python=["3.12"])
def lint(session: nox.Session):
    session.run("uv", "sync", "--no-dev", "--group", "lint")
    session.run("uv", "run", "ruff", "check", "src")


@nox.session(python=["3.12"])
def type_check(session: nox.Session):
    session.run("uv", "sync", "--no-dev", "--group", "type")
    session.run("uv", "run", "mypy", "src")


@nox.session(python=["3.10", "3.11", "3.12", "3.13"])
def test(session: nox.Session):
    session.run("uv", "sync", "--no-dev", "--group", "test")
    session.run(
        "uv", "run", "pytest", "--cov-branch", "--cov-report=xml", "-n", "auto"
    )
